# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-04-06 15:30:50
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-07-05 12:22:37

from __future__ import print_function, division, absolute_import
import os
import pytest
import requests
from flask import url_for
from selenium import webdriver
from marvin.web import create_app
from marvin.tests.web.frontend.live_server import live_server
from marvin.web.settings import TestConfig, CustomConfig

browserstack = os.environ.get('USE_BROWSERSTACK', None)

if browserstack:
    osdict = {'OS X': ['El Capitan', 'Sierra']}
    browserdict = {'chrome': ['55', '54'], 'firefox': ['52', '51'], 'safari': ['10', '9.1']}
else:
    osdict = {'OS X': ['El Capitan']}
    browserdict = {'chrome': ['55']}

osstuff = [(k, i) for k, v in osdict.items() for i in v]
browserstuff = [(k, i) for k, v in browserdict.items() for i in v]


@pytest.fixture(params=osstuff)
def osinfo(request):
    return request.param


@pytest.fixture(params=browserstuff)
def browserinfo(request):
    return request.param


@pytest.fixture(scope='session')
def app():
    object_config = type('Config', (TestConfig, CustomConfig), dict())
    app = create_app(debug=True, local=True, object_config=object_config)
    # app = create_app(debug=True, local=True, use_profiler=False)
    # app.config['TESTING'] = True
    # app.config['WTF_CSRF_ENABLED'] = False
    # app.config['PRESERVE_CONTEXT_ON_EXCEPTION'] = False
    app.config['LIVESERVER_PORT'] = 8943
    return app


@pytest.fixture(scope='function')
def base_url(live_server):
    return '{0}/marvin2/'.format(live_server.url())


@pytest.fixture(scope='function')
def driver(base_url, osinfo, browserinfo):
    ostype, os_version = osinfo
    browser, browser_version = browserinfo
    buildid = '{0}_{1}_{2}_{3}'.format(ostype.lower().replace(' ', '_'),
                                       os_version.lower().replace(' ', '_'), browser, browser_version)
    # skip some combinations
    if os_version == 'Sierra' and browser == 'safari' and browser_version == '9.1':
        pytest.skip('cannot have Sierra running safari 9.1')
    elif os_version == 'El Capitan' and browser == 'safari' and browser_version == '10':
        pytest.skip('cannot have El Capitan running safari 10')

    # set driver
    if browserstack:
        username = os.environ.get('BROWSERSTACK_USER', None)
        access = os.environ.get('BROWSERSTACK_ACCESS_KEY', None)
        desired_cap = {'os': ostype, 'os_version': os_version, 'browser': browser, 'build': buildid,
                       'browser_version': browser_version, 'project': 'marvin', 'resolution': '1920x1080'}
        desired_cap['browserstack.local'] = True
        desired_cap['browserstack.debug'] = True
        desired_cap['browserstack.localIdentifier'] = os.environ['BROWSERSTACK_LOCAL_IDENTIFIER']
        driver = webdriver.Remote(
            command_executor='http://{0}:{1}@hub.browserstack.com:80/wd/hub'.format(username, access),
            desired_capabilities=desired_cap)
    else:
        if browser == 'chrome':
            driver = webdriver.Chrome()
        elif browser == 'firefox':
            driver = webdriver.Firefox()
        elif browser == 'safari':
            driver = webdriver.Safari()

    driver.get(base_url)
    yield driver
    # teardown
    driver.quit()


@pytest.mark.usefixtures('live_server')
class TestLiveServer(object):

    def test_server_is_up_and_running(self, base_url):
        response = requests.get(base_url)
        assert response.status_code == 200







from __future__ import print_function

from os import getenv, makedirs
from os.path import exists, join, basename
from math import log10
from sys import exit
from subprocess import Popen, STDOUT
from shlex import split
from tempfile import TemporaryFile
from time import time, sleep
from glob import iglob
from datetime import datetime

class Cli:
    """Class for providing command line interface (cli) sync scripts, and logs to local disk
    """
    
    tmp_dir = '/tmp'

    def __init__(self, label=None, data_dir=None, verbose=False):
        self.label = label if label else 'sdss_access'
        self.data_dir = data_dir if data_dir else getenv('SDSS_ACCESS_DATA_DIR')
        self.ready = exists(self.data_dir) if self.data_dir else False
        if not self.ready and exists(self.tmp_dir): self.data_dir = self.tmp_dir
        self.now = datetime.now().strftime("%Y%m%d")
        self.env = None
        self.verbose = verbose

    def set_dir(self):
        if exists(self.data_dir) and self.label:
            label_dir = join(self.data_dir,self.label)
            position = len(self.now)+1
            maxnumber = 0
            for dir in iglob(join(label_dir,"{now}_*".format(now=self.now))):
                dirname = basename(dir)
                try: dirnumber = int(dirname[position:])
                except: dirnumber = 0
                maxnumber = max(maxnumber,dirnumber)
            self.dir = join(label_dir,"{now}_{number:03d}".format(now=self.now,number=maxnumber+1))
            if not exists(self.dir):
                if self.verbose: print("SDSS_ACCESS> CREATE {dir}".format(dir=self.dir))
                makedirs(self.dir)
        else: self.dir = None
            
    def get_path(self,index=None):
        try: index = int(index)
        except: index = 0
        return join(self.dir,"{label}_{index:02d}".format(label=self.label,index=index)) if self.dir and self.label else None
    
    def write_lines(self, path=None, lines=None):
        if path and lines:
            with open(path,'w') as file:
                file.write("\n".join(lines)+"\n")

    def get_background_process(self, command=None, logfile=None, errfile=None, pause=1):
        if command:
            stdout = logfile if logfile else STDOUT
            stderr = errfile if errfile else STDOUT
            background_process = Popen(split(str(command)), env=self.env, stdout=stdout, stderr=stderr)
            if pause: sleep(pause)
        else: background_process = None
        return background_process

    def wait_for_processes(self, processes, pause=60):
        print("SDSS_ACCESS> syncing... please wait")
        while any([process.poll() is None for process in processes]): sleep(pause)
        print("SDSS_ACCESS> Done!")
        self.returncode = tuple([process.returncode for process in processes])

    def foreground_run(self, command, test=False, logger=None, logall=False, message=None, outname=None, errname=None):
        """A convenient wrapper to log and perform system calls.

        Parameters
        ----------
        command : str
            The system command to run.  It will be split internally by shlex.split().
        test : bool, optional
            Set this to true to not actually run the commands.
        logger : logging.logging, optional
            If passed a logging object, diagnostic output will be set to this object.
        logall : bool, optional
            Set this to true to always log stdout and stderr (at level DEBUG).
            Otherwise stdout and stderr will only be logged for nonzero status
        message : str, optional
            Call logger.critical(message) with this message and then call
            sys.exit(status).
        outname : str, optional
            If set, use ``outname`` as the name of a file to contain stdout.  It
            is the responsibility of the caller of this function to clean up
            the resulting file.  Otherwise a temporary file will be used.
        errname : str, optional
            Same as ``outname``, but will contain stderr.

        Returns
        -------
        (status,out,err) : tuple
            The exit status, stdout and stderr of the process

        Examples
        --------
        >>> status,out,err = transfer.common.system_call('date')
        """
        if logger is not None:
            logger.debug(command)
        status = 0
        out = ''
        err = ''
        if not test:
            if outname is None:
                outfile = TemporaryFile()
            else:
                outfile = open(outname,'w+')
            if errname is None:
                errfile = TemporaryFile()
            else:
                errfile = open(errname,'w+')
            proc = Popen(split(str(command)),stdout=outfile,stderr=errfile,env=self.env)
            tstart = time()
            while proc.poll() is None:
                elapsed = time() - tstart
                if elapsed > 500000:
                    message = "Process still running after more than 5 days!"
                    proc.kill()
                    break
                tsleep = 10**(int(log10(elapsed))-1)
                if tsleep < 1:
                    tsleep = 1
                sleep(tsleep)
            # proc.wait()
            status = proc.returncode
            outfile.seek(0)
            out = outfile.read()
            errfile.seek(0)
            err = errfile.read()
            outfile.close()
            errfile.close()
            if logger is not None:
                if status == 0 and logall:
                    if len(out) > 0:
                        logger.debug('STDOUT = \n' + out)
                    if len(err) > 0:
                        logger.debug('STDERR = \n' + err)
                if status != 0:
                    logger.error('status = {0}'.format(status))
                    if len(out) > 0:
                        logger.error('STDOUT = \n' + out)
                    if len(err) > 0:
                        logger.error('STDERR = \n' + err)
                    if message is not None:
                        logger.critical(message)
                        exit(status)
        return (status,out,err)

class CliError(Exception):
    pass





# Copyright 2015 SolidBuilds.com. All rights reserved.
#
# Authors: Ling Thio <ling.thio@gmail.com>

from __future__ import print_function
from sqlalchemy_boolean_search import parse_boolean_search
from .models import Record, Parent, GrandParent
import pytest


def test_field_names(db):
    # Create records
    grandparent = GrandParent(name='GrandParent')
    parent = Parent(name='Parent', grandparent=grandparent)
    record = Record(string='Record', parent=parent)
    db.session.add(record)
    db.session.commit()

    # Test non-hierarchical name
    expression = parse_boolean_search('string==Record')
    record = Record.query.filter(expression.filter(Record)).first()
    assert record is not None
    assert record.string == 'Record'


@pytest.mark.skip(reason="tests no longer works for relationships")
def test_fields_across_relationships(db):
    # Create records
    grandparent = GrandParent(name='GrandParent')
    parent = Parent(name='Parent', grandparent=grandparent)
    record = Record(string='Record', parent=parent)
    db.session.add(record)
    db.session.commit()

    # Test level-1 hierarchy name
    expression = parse_boolean_search('parent.name==Parent')
    record = Record.query.filter(expression.filter(Record)).first()
    assert record is not None
    assert record.string == 'Record'

    # Test level-2 hierarchy name
    expression = parse_boolean_search('parent.grandparent.name==GrandParent')
    record = Record.query.filter(expression.filter(Record)).first()
    assert record is not None
    assert record.string == 'Record'

    # Delete records
    db.session.delete(record)
    db.session.commit()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module defines a logging class based on the built-in logging module.
The module is heavily based on the astropy logging system.
"""

from __future__ import print_function

import os
import sys
import logging
from logging import FileHandler
from logging.handlers import TimedRotatingFileHandler
import warnings
from brain.core.colourPrint import colourPrint
import shutil
import re
from textwrap import TextWrapper
from brain.core.exceptions import BrainError


# Initialize by calling initLog()
log = None

# Adds custom log level for important messages
IMPORTANT = 25
logging.addLevelName(IMPORTANT, 'IMPORTANT')

ansi_escape = re.compile(r'\x1b[^m]*m')


def important(self, message, *args, **kws):
    self._log(IMPORTANT, message, args, **kws)

logging.Logger.important = important


def initLog(logFilePath, logLevel='INFO', logFileLevel='DEBUG',
            mode='append', wrapperLength=80):

    logging.setLoggerClass(BrainLogger)
    log = logging.getLogger('Brain')
    log._set_defaults(logLevel=logLevel, logFileLevel=logFileLevel, logFilePath=logFilePath,
                      mode=mode, wrapperLength=wrapperLength)

    return log


class MyFormatter(logging.Formatter):

    warning_fmp = '%(asctime)s - %(levelname)s: %(message)s [%(origin)s]'
    info_fmt = '%(asctime)s - %(levelname)s - %(message)s [%(funcName)s @ ' + \
        '%(filename)s]'

    def __init__(self, fmt='%(levelname)s - %(message)s [%(funcName)s @ ' +
                 '%(filename)s]'):
        logging.Formatter.__init__(self, fmt, datefmt='%Y-%m-%d %H:%M:%S')

    def format(self, record):

        # Save the original format configured by the user
        # when the logger formatter was instantiated
        format_orig = self._fmt

        # Replace the original format with one customized by logging level
        if record.levelno == logging.DEBUG:
            self._fmt = MyFormatter.info_fmt

        elif record.levelno == logging.INFO:
            self._fmt = MyFormatter.info_fmt

        elif record.levelno == logging.ERROR:
            self._fmt = MyFormatter.info_fmt

        elif record.levelno == logging.WARNING:
            self._fmt = MyFormatter.warning_fmp

        elif record.levelno == IMPORTANT:
            self._fmt = MyFormatter.info_fmt

        record.msg = ansi_escape.sub('', record.msg)

        # Call the original formatter class to do the grunt work
        result = logging.Formatter.format(self, record)

        # Restore the original format configured by the user
        self._fmt = format_orig

        return result


Logger = logging.getLoggerClass()
fmt = MyFormatter()


class BrainLogger(Logger):
    """This class is used to set up the logging system.

    The main functionality added by this class over the built-in
    logging.Logger class is the ability to keep track of the origin of the
    messages, the ability to enable logging of warnings.warn calls and
    exceptions, and the addition of colorized output and context managers to
    easily capture messages to a file or list.

    """

    def saveLog(self, path):
        shutil.copyfile(self.logFilename, os.path.expanduser(path))

    def _showwarning(self, *args, **kwargs):

        warning = args[0]
        message = '{0}: {1}'.format(warning.__class__.__name__, args[0])
        mod_path = args[2]

        mod_name = None
        mod_path, ext = os.path.splitext(mod_path)
        for name, mod in list(sys.modules.items()):
            path = os.path.splitext(getattr(mod, '__file__', ''))[0]
            if path == mod_path:
                mod_name = mod.__name__
                break

        self.warning(message, extra={'origin': mod_name or ''})

    def warning(self, message, *args, **kwargs):

        # if there are args then we have old style formating
        if args:
            message = message % args

        # check for the extra keyword in kwargs
        extra = kwargs.get('extra', None)
        if extra is None:
            extra = {'origin': ''}

        super(BrainLogger, self).warning(message, extra=extra)

    def _stream_formatter(self, record):
        """The formatter for standard output."""
        if record.levelno < logging.DEBUG:
            print(record.levelname, end='')
        elif(record.levelno < logging.INFO):
            colourPrint(record.levelname, 'green', end='')
        elif(record.levelno < IMPORTANT):
            colourPrint(record.levelname, 'magenta', end='')
        elif(record.levelno < logging.WARNING):
            colourPrint(record.levelname, 'lightblue', end='')
        elif(record.levelno < logging.ERROR):
            colourPrint(record.levelname, 'brown', end='')
        else:
            colourPrint(record.levelname, 'red', end='')

        if record.levelno == logging.WARN:
            message = '{0}'.format(record.msg[record.msg.find(':') + 2:])
        else:
            message = '{0}'.format(record.msg)

        if len(message) > self.wrapperLength:
            tw = TextWrapper()
            tw.width = self.wrapperLength
            tw.subsequent_indent = ' ' * (len(record.levelname) + 2)
            tw.break_on_hyphens = False
            message = '\n'.join(tw.wrap(message))

        print(': ' + message)

    def _set_defaults(self,
                      logLevel='WARNING',
                      logFileLevel='INFO',
                      logFilePath='~/.brain/brain.log',
                      mode='append',
                      wrapperLength=70):
        """Reset logger to its initial state."""

        # Remove all previous handlers
        for handler in self.handlers[:]:
            self.removeHandler(handler)

        # Set levels
        self.setLevel('DEBUG')

        # Set up the stdout handler
        self.sh = logging.StreamHandler()
        self.sh.emit = self._stream_formatter
        self.addHandler(self.sh)

        self.wrapperLength = wrapperLength

        # Set up the main log file handler if requested (but this might fail if
        # configuration directory or log file is not writeable).

        logFilePath = os.path.expanduser(logFilePath)
        logDir = os.path.dirname(logFilePath)
        if not os.path.exists(logDir):
            os.mkdir(logDir)

        try:
            if mode.lower() == 'overwrite':
                self.fh = FileHandler(logFilePath, mode='w')
            elif mode.lower() == 'append':
                self.fh = TimedRotatingFileHandler(
                    logFilePath, when='midnight', utc=True)
            else:
                raise BrainError('logger mode {0} not recognised'.format(mode))
        except (IOError, OSError) as e:
            warnings.warn(
                'log file {0!r} could not be opened for writing: '
                '{1}'.format(logFilePath, unicode(e)), RuntimeWarning)
        else:
            self.fh.setFormatter(fmt)
            self.addHandler(self.fh)

        # Adds a header only to the file handler
        self.sh.setLevel(logging.CRITICAL)
        self.fh.setLevel(logging.DEBUG)
        self.debug('')
        self.debug('--------------------------------')
        self.debug('----- Restarting logger. -------')
        self.debug('--------------------------------')

        self.sh.setLevel(logLevel)
        self.fh.setLevel(logFileLevel)

        self.logFilename = logFilePath
        warnings.showwarning = self._showwarning



#!/usr/bin/env python
# encoding: utf-8
#
# test_colorbar.py
#
# @Author: Brett Andrews <andrews>
# @Date:   2017-07-26 11:07:00
# @Last modified by:   andrews
# @Last modified time: 2017-07-30 00:07:14

import os

import pytest
import numpy as np
import astropy

from marvin.utils.plot import colorbar

@pytest.fixture(scope='function')
def image():
    data = np.linspace(1, 9, 9).reshape(3, 3)
    mask = np.array([[0, 1, 0], [1, 0, 1], [0, 1, 0]])
    masked_array = np.ma.array(data, mask=mask)
    return masked_array


class TestColorMap(object):

    def test_linearlab_filename_exists(self):
        assert os.path.isfile(colorbar._linearlab_filename())


class TestRange(object):

    @pytest.mark.parametrize('cbrange, expected',
                             [([1.5, 1.75], []),
                              ([1.5, 4], [2, 3]),
                              ([1.5, 8], [2, 3, 6]),
                              ([1.5, 18], [2, 3, 6, 10]),
                              ([1.5, 45], [2, 3, 6, 10, 20, 30]),
                              ([2, 8], [2, 3, 6]),
                              ([3, 25], [3, 6, 10, 20])
                              ])
    def test_log_cbticks(self, cbrange, expected):
        assert (colorbar._log_cbticks(cbrange) == expected).all()
        assert (colorbar._set_cbticks(cbrange, {'log_cb': True})[1] == expected).all()

    @pytest.mark.parametrize('value, expected',
                             [(1, '1'),
                              (10, '10'),
                              (100, '100'),
                              (150, '150'),
                              (1000, '1e3'),
                              (1500, '1e3'),
                              (10000, '1e4'),
                              (0.1, '0.1'),
                              (0.11, '0.1')])
    def test_log_tick_format(self, value, expected):
        assert colorbar._log_tick_format(value) == expected
    
    @pytest.mark.parametrize('d, cbrange',
                             [({'vmin': 0, 'vmax': 1}, None),
                              ({'vmax': 1}, [0, 2]),
                              ({'vmin': 0}, [0.1, 1]),
                              ({}, [0, 1])
                              ])
    def test_set_vmin_vmax(self, d, cbrange):
        expected = {'vmin': 0, 'vmax': 1}
        assert colorbar._set_vmin_vmax(d, cbrange) == expected

    @pytest.mark.parametrize('image, sigma, expected',
                             [(np.ma.array([1, 3, 7, 13, 29, 35], mask=[0, 1, 1, 0, 0, 0]), 1,
                               [13, 29])])
    def test_cbrange_sigma_clip(self, image, sigma, expected):
        assert colorbar._cbrange_sigma_clip(image, sigma=sigma) == expected







#!/usr/bin/env python
# encoding: utf-8
"""

core.py

Licensed under a 3-clause BSD license.

Revision history:
    12 Feb 2016 J. Sánchez-Gallego
      Initial version

"""

from __future__ import division
from __future__ import print_function

import os
import warnings

import astropy.io.fits

from brain.core.exceptions import BrainError
from collections import OrderedDict
import marvin
import marvin.api.api
from marvin.core import marvin_pickle

from marvin.core.exceptions import MarvinUserWarning, MarvinError
from marvin.core.exceptions import MarvinMissingDependency, MarvinBreadCrumb
from marvin.utils.db import testDbConnection
from marvin.utils.general import mangaid2plateifu, get_nsa_data

try:
    from sdss_access.path import Path
except ImportError:
    Path = None

try:
    from sdss_access import RsyncAccess
except ImportError:
    RsyncAccess = None


__all__ = ['MarvinToolsClass', 'Dotable', 'DotableCaseInsensitive']


def kwargsGet(kwargs, key, replacement):
    """As kwargs.get but uses replacement if the value is None."""

    if key not in kwargs:
        return replacement
    elif key in kwargs and kwargs[key] is None:
        return replacement
    else:
        return kwargs[key]


breadcrumb = MarvinBreadCrumb()


# TODO: probably MarvinToolsClass should be derived from an AbstractClass

class MarvinToolsClass(object):

    def __init__(self, *args, **kwargs):
        """Marvin tools main super class.

        This super class implements the decision tree for using local files,
        database, or remote connection when initialising a Marvin tools
        object.

        """

        self.data = kwargsGet(kwargs, 'data', None)

        self.filename = kwargsGet(kwargs, 'filename', None)
        if self.filename:
            self.filename = os.path.realpath(os.path.expanduser(self.filename))

        self.mangaid = kwargsGet(kwargs, 'mangaid', None)
        self.plateifu = kwargsGet(kwargs, 'plateifu', None)

        self.mode = kwargsGet(kwargs, 'mode', marvin.config.mode)

        self._release = kwargsGet(kwargs, 'release', marvin.config.release)

        self._drpver, self._dapver = marvin.config.lookUpVersions(release=self._release)
        self._drpall = kwargsGet(kwargs, 'drpall', marvin.config._getDrpAllPath(self._drpver))

        self._nsa = None
        self.nsa_source = kwargs.pop('nsa_source', 'auto')
        assert self.nsa_source in ['auto', 'nsa', 'drpall'], \
            'nsa_source must be one of auto, nsa, or drpall'

        self._forcedownload = kwargsGet(kwargs, 'download', marvin.config.download)

        self.data_origin = None

        args = [self.filename, self.plateifu, self.mangaid]
        assert any(args), 'Enter filename, plateifu, or mangaid!'

        if self.filename:
            self.plateifu = None
            self.mangaid = None
        elif self.plateifu:
            self.filename = None
            self.mangaid = None
        elif self.mangaid:
            self.filename = None
            self.plateifu = mangaid2plateifu(self.mangaid,
                                             drpall=self._drpall,
                                             drpver=self._drpver)

        # drop breadcrumb
        breadcrumb.drop(message='Initializing MarvinTool {0}'.format(self.__class__),
                        category=self.__class__)

        if self.mode == 'local':
            self._doLocal()
        elif self.mode == 'remote':
            self._doRemote()
        elif self.mode == 'auto':
            try:
                self._doLocal()
            except Exception as ee:
                if self.filename:
                    # If the input contains a filename we don't want to go into remote mode.
                    raise(ee)
                else:
                    warnings.warn('local mode failed. Trying remote now.', MarvinUserWarning)
                    self._doRemote()

        # Sanity check to make sure data_origin has been properly set.
        assert self.data_origin in ['file', 'db', 'api'], 'data_origin is not properly set.'

    def _doLocal(self):
        """Tests if it's possible to load the data locally."""

        if self.filename:

            if os.path.exists(self.filename):
                self.mode = 'local'
                self.data_origin = 'file'
            else:
                raise MarvinError('input file {0} not found'.format(self.filename))

        elif self.plateifu:

            testDbConnection(marvin.marvindb.session)

            if marvin.marvindb.db:
                self.mode = 'local'
                self.data_origin = 'db'
            else:
                # TODO - fix verbosity later, check for more advanced db failures
                # warnings.warn('DB connection failed with error: {0}.'.format(dbStatus['error']),
                #               MarvinUserWarning)

                fullpath = self._getFullPath()

                if fullpath and os.path.exists(fullpath):
                    self.mode = 'local'
                    self.filename = fullpath
                    self.data_origin = 'file'
                else:
                    if self._forcedownload:
                        self.download()
                        # raise NotImplementedError('sdsssync not yet implemented')
                        self.data_origin = 'file'
                        # When implemented, this should download the data and
                        # then kwargs['filename'] = downloaded_path and
                        # kwargs['mode'] = local
                    else:
                        raise MarvinError('this is the end of the road. '
                                          'Try using some reasonable inputs.')

    def _doRemote(self):
        """Tests if remote connection is possible."""

        if self.filename:
            raise MarvinError('filename not allowed in remote mode.')
        else:
            self.mode = 'remote'
            self.data_origin = 'api'

    def download(self, pathType=None, **pathParams):
        ''' Download using sdss_access Rsync '''
        if not RsyncAccess:
            raise MarvinError('sdss_access is not installed')
        else:
            rsync_access = RsyncAccess()
            rsync_access.remote()
            rsync_access.add(pathType, **pathParams)
            rsync_access.set_stream()
            rsync_access.commit()
            paths = rsync_access.get_paths()
            self.filename = paths[0]  # doing this for single files, may need to change

    def _getFullPath(self, pathType=None, url=None, **pathParams):
        """Returns the full path of the file in the tree."""

        if not Path:
            raise MarvinMissingDependency('sdss_access is not installed')
        else:
            try:
                if url:
                    fullpath = Path().url(pathType, **pathParams)
                else:
                    fullpath = Path().full(pathType, **pathParams)
            except Exception as ee:
                warnings.warn('sdss_access was not able to retrieve the full path of the file. '
                              'Error message is: {0}'.format(str(ee)), MarvinUserWarning)
                fullpath = None

        return fullpath

    def _toolInteraction(self, url, params=None):
        """Runs an Interaction and passes self._release."""

        params = params or {'release': self._release}
        return marvin.api.api.Interaction(url, params=params)

    def __getstate__(self):

        if self.data_origin == 'db':
            raise MarvinError('objects with data_origin=\'db\' cannot be saved.')

        odict = self.__dict__.copy()
        del odict['data']

        return odict

    def __setstate__(self, idict):

        data = None
        if idict['data_origin'] == 'file':
            try:
                data = astropy.io.fits.open(idict['filename'])
            except Exception as ee:
                warnings.warn('there was a problem reloading the FITS object: {0}. '
                              'The object has been unpickled but not all the functionality '
                              'will be available.'.format(str(ee)), MarvinUserWarning)

        self.__dict__.update(idict)
        self.data = data

    def save(self, path=None, overwrite=False):
        """Pickles the object.

        If ``path=None``, uses the default location of the file in the tree
        but changes the extension of the file to ``.mpf``. Returns the path
        of the saved pickle file.

        Parameters:
            obj:
                Marvin object to pickle.
            path (str):
                Path of saved file. Default is ``None``.
            overwrite (bool):
                If ``True``, overwrite existing file. Default is ``False``.

        Returns:
            str:
                Path of saved file.
        """

        return marvin_pickle.save(self, path=path, overwrite=overwrite)

    @classmethod
    def restore(cls, path, delete=False):
        """Restores a MarvinToolsClass object from a pickled file.

        If ``delete=True``, the pickled file will be removed after it has been
        unplickled. Note that, for objects with ``data_origin='file'``, the
        original file must exists and be in the same path as when the object
        was first created.

        """

        return marvin_pickle.restore(path, delete=delete)

    @property
    def nsa(self):
        """Returns the contents of the NSA catalogue for this target."""

        if self._nsa is None:

            if self.nsa_source == 'auto':
                if self.data_origin == 'file':
                    nsa_source = 'drpall'
                else:
                    nsa_source = 'nsa'
            else:
                nsa_source = self.nsa_source

            try:
                self._nsa = get_nsa_data(self.mangaid, mode='auto',
                                         source=nsa_source,
                                         drpver=self._drpver,
                                         drpall=self._drpall)
            except (MarvinError, BrainError):
                warnings.warn('cannot load NSA information for mangaid={0}.'.format(self.mangaid))
                return None

        return self._nsa

    @property
    def release(self):
        """Returns the release."""

        return self._release

    @release.setter
    def release(self, value):
        """Fails when trying to set the release after instatiation."""

        raise MarvinError('the release cannot be changed once the object has been instantiated.')


class Dotable(dict):
    """A custom dict class that allows dot access to nested dictionaries.

    Copied from http://hayd.github.io/2013/dotable-dictionaries/. Note that
    this allows you to use dots to get dictionary values, but not to set them.

    """

    def __getattr__(self, value):
        if '__' in value:
            return dict.__getattr__(self, value)
        else:
            return self.__getitem__(value)

    # def __init__(self, d):
    #     dict.__init__(self, ((k, self.parse(v)) for k, v in d.iteritems()))

    @classmethod
    def parse(cls, v):
        if isinstance(v, dict):
            return cls(v)
        elif isinstance(v, list):
            return [cls.parse(i) for i in v]
        else:
            return v


class DotableCaseInsensitive(Dotable):
    """Like dotable but access to attributes and keys is case insensitive."""

    def _match(self, list_of_keys, value):

        lower_values = [str(xx).lower() for xx in list_of_keys]
        if value.lower() in lower_values:
            return list_of_keys[lower_values.index(value.lower())]
        else:
            return False

    def __getattr__(self, value):
        if '__' in value:
            return super(DotableCaseInsensitive, self).__getattr__(value)
        return self.__getitem__(value)

    def __getitem__(self, value):
        key = self._match(list(self.keys()), value)
        if key is False:
            raise KeyError('{0} key or attribute not found'.format(value))
        return dict.__getitem__(self, key)


class OrderedDefaultDict(OrderedDict):
    def __init__(self, default_factory=None, *args, **kwargs):
        OrderedDict.__init__(self, *args, **kwargs)
        self.default_factory = default_factory

    def __missing__(self, key):
        result = self[key] = self.default_factory()
        return result



from flask_classy import route
from flask import request, jsonify, Response, current_app
from marvin.tools.query import doQuery, Query
from marvin.core.exceptions import MarvinError
from marvin.api.base import BaseView, arg_validate as av
from marvin.utils.db import get_traceback
from marvin.tools.query.query_utils import bestparams
from marvin.web.extensions import limiter
import json


def _getCubes(searchfilter, **kwargs):
    """Run query locally at Utah."""

    release = kwargs.pop('release', None)
    kwargs['returnparams'] = kwargs.pop('params', None)
    kwargs['returntype'] = kwargs.pop('rettype', None)

    try:
        # q, r = doQuery(searchfilter=searchfilter, returnparams=params, release=release,
        #                mode='local', returntype=rettype, limit=limit, order=order, sort=sort)
        q, r = doQuery(searchfilter=searchfilter, release=release, **kwargs)
    except Exception as e:
        raise MarvinError('Query failed with {0}: {1}'.format(e.__class__.__name__, e))

    results = r.results

    # get the subset keywords
    start = kwargs.get('start', None)
    end = kwargs.get('end', None)
    limit = kwargs.get('limit', None)
    params = kwargs.get('params', None)

    # get a subset
    chunk = None
    if start:
        chunk = int(end) - int(start)
        results = r.getSubset(int(start), limit=chunk)
    chunk = limit if not chunk else limit
    runtime = {'days': q.runtime.days, 'seconds': q.runtime.seconds, 'microseconds': q.runtime.microseconds}
    output = dict(data=results, query=r.showQuery(), chunk=limit,
                  filter=searchfilter, params=q.params, returnparams=params, runtime=runtime,
                  queryparams_order=q.queryparams_order, count=len(results), totalcount=r.totalcount)
    return output


class QueryView(BaseView):
    """Class describing API calls related to queries."""
    decorators = [limiter.limit("60/minute")]

    def index(self):
        '''Returns general query info

        .. :quickref: Query; Get general query info

        :query string release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson string data: data message
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin2/api/query/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": "this is a query!"
           }

        '''
        self.results['data'] = 'this is a query!'
        self.results['status'] = 1
        return jsonify(self.results)

    @route('/cubes/', methods=['GET', 'POST'], endpoint='querycubes')
    @av.check_args(use_params='query', required='searchfilter')
    def cube_query(self, args):
        ''' Performs a remote query

        .. :quickref: Query; Perform a remote query

        :query string release: the release of MaNGA
        :form searchfilter: your string searchfilter expression
        :form params: the list of return parameters
        :form rettype: the string indicating your Marvin Tool conversion object
        :form limit: the limiting number of results to return for large results
        :form sort: a string parameter name to sort on
        :form order: the order of the sort, either ``desc`` or ``asc``
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson string data: dictionary of returned data
        :json list results: the list of results
        :json string query: the raw SQL string of your query
        :json int chunk: the page limit of the results
        :json string filter: the searchfilter used
        :json list returnparams: the list of return parameters
        :json list params: the list of parameters used in the query
        :json list queryparams_order: the list of parameters used in the query
        :json dict runtime: a dictionary of query time (days, minutes, seconds)
        :json int totalcount: the total count of results
        :json int count: the count in the current page of results
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin2/api/query/cubes/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5", "searchfilter": "nsa.z<0.1"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "chunk": 100,
              "count": 4,
              "data": [["1-209232",8485,"8485-1901","1901",0.0407447],
                       ["1-209113",8485,"8485-1902","1902",0.0378877],
                       ["1-209191",8485,"8485-12701","12701",0.0234253],
                       ["1-209151",8485,"8485-12702","12702",0.0185246]
                      ],
              "filter": "nsa.z<0.1",
              "params": ["cube.mangaid","cube.plate","cube.plateifu","ifu.name","nsa.z"],
              "query": "SELECT ... FROM ... WHERE ...",
              "queryparams_order": ["mangaid","plate","plateifu","name","z"],
              "returnparams": null,
              "runtime": {"days": 0,"microseconds": 55986,"seconds": 0},
              "totalcount": 4
           }

        '''
        searchfilter = args.pop('searchfilter', None)
        # searchfilter = self.results['inconfig'].get('searchfilter', None)
        # params = self.results['inconfig'].get('params', None)
        # rettype = self.results['inconfig'].get('returntype', None)
        # limit = self.results['inconfig'].get('limit', 100)
        # sort = self.results['inconfig'].get('sort', None)
        # order = self.results['inconfig'].get('order', 'asc')
        # release = self.results['inconfig'].get('release', None)

        try:
            # res = _getCubes(searchfilter, params=params, rettype=rettype,
            #                 limit=limit, sort=sort, order=order, release=release)
            res = _getCubes(searchfilter, **args)
        except MarvinError as e:
            self.results['error'] = str(e)
            self.results['traceback'] = get_traceback(asstring=True)
        else:
            self.results['status'] = 1
            self.update_results(res)

        # this needs to be json.dumps until sas-vm at Utah updates to 2.7.11
        return Response(json.dumps(self.results), mimetype='application/json')

    @route('/cubes/getsubset/', methods=['GET', 'POST'], endpoint='getsubset')
    @av.check_args(use_params='query', required=['searchfilter', 'start', 'end'])
    def query_getsubset(self, args):
        ''' Remotely grab a subset of results from a query

        .. :quickref: Query; Grab a subset of results from a remote query

        :query string release: the release of MaNGA
        :form searchfilter: your string searchfilter expression
        :form params: the list of return parameters
        :form rettype: the string indicating your Marvin Tool conversion object
        :form start: the starting page index of results you wish to grab
        :form end: the ending page index of the results you wish to grab
        :form limit: the limiting number of results to return for large results
        :form sort: a string parameter name to sort on
        :form order: the order of the sort, either ``desc`` or ``asc``
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson string data: dictionary of returned data
        :json list results: the list of results
        :json string query: the raw SQL string of your query
        :json int chunk: the page limit of the results
        :json string filter: the searchfilter used
        :json list returnparams: the list of return parameters
        :json list params: the list of parameters used in the query
        :json list queryparams_order: the list of parameters used in the query
        :json dict runtime: a dictionary of query time (days, minutes, seconds)
        :json int totalcount: the total count of results
        :json int count: the count in the current page of results
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin2/api/query/cubes/getsubset/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5", "searchfilter": "nsa.z<0.1", "start":10, "end":15},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "chunk": 100,
              "count": 4,
              "data": [["1-209232",8485,"8485-1901","1901",0.0407447],
                       ["1-209113",8485,"8485-1902","1902",0.0378877],
                       ["1-209191",8485,"8485-12701","12701",0.0234253],
                       ["1-209151",8485,"8485-12702","12702",0.0185246]
                      ],
              "filter": "nsa.z<0.1",
              "params": ["cube.mangaid","cube.plate","cube.plateifu","ifu.name","nsa.z"],
              "query": "SELECT ... FROM ... WHERE ...",
              "queryparams_order": ["mangaid","plate","plateifu","name","z"],
              "returnparams": null,
              "runtime": {"days": 0,"microseconds": 55986,"seconds": 0},
              "totalcount": 4
           }

        '''
        searchfilter = args.pop('searchfilter', None)
        # searchfilter = self.results['inconfig'].get('searchfilter', None)
        # params = self.results['inconfig'].get('params', None)
        # start = self.results['inconfig'].get('start', None)
        # end = self.results['inconfig'].get('end', None)
        # rettype = self.results['inconfig'].get('returntype', None)
        # limit = self.results['inconfig'].get('limit', 100)
        # sort = self.results['inconfig'].get('sort', None)
        # order = self.results['inconfig'].get('order', 'asc')
        # release = self.results['inconfig'].get('release', None)

        try:
            # res = _getCubes(searchfilter, params=params, start=int(start),
            #                 end=int(end), rettype=rettype, limit=limit,
            #                 sort=sort, order=order, release=release)
            res = _getCubes(searchfilter, **args)
        except MarvinError as e:
            self.results['error'] = str(e)
            self.results['traceback'] = get_traceback(asstring=True)
        else:
            self.results['status'] = 1
            self.update_results(res)

        # this needs to be json.dumps until sas-vm at Utah updates to 2.7.11
        return Response(json.dumps(self.results), mimetype='application/json')

    @route('/getparamslist/', methods=['GET', 'POST'], endpoint='getparams')
    @av.check_args(use_params='query', required='paramdisplay')
    def getparamslist(self, args):
        ''' Retrieve a list of all available input parameters into the query

        .. :quickref: Query; Get a list of all or "best" queryable parameters

        :query string release: the release of MaNGA
        :form paramdisplay: ``all`` or ``best``, type of parameters to return
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson string data: dictionary of returned data
        :json list params: the list of queryable parameters
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin2/api/query/getparamslist/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": ['nsa.z', 'cube.ra', 'cube.dec', ...]
           }

        '''
        paramdisplay = args.pop('paramdisplay', 'all')
        q = Query(mode='local', release=args['release'])
        if paramdisplay == 'all':
            params = q.get_available_params('all')
        elif paramdisplay == 'best':
            params = bestparams
        self.results['data'] = params
        self.results['status'] = 1
        output = jsonify(self.results)
        return output

    @route('/cleanup/', methods=['GET', 'POST'], endpoint='cleanupqueries')
    @av.check_args(use_params='query', required='task')
    def cleanup(self, args):
        ''' Clean up idle server-side queries or retrieve the list of them

        Do not use!

        .. :quickref: Query; Send a cleanup command to the server-side database

        :query string release: the release of MaNGA
        :form task: ``clean`` or ``getprocs``, the type of task to run
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson string data: dictionary of returned data
        :json string clean: clean success message
        :json list procs: the list of processes currently running on the db
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin2/api/query/cleanup/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": 'clean success'
           }

        '''
        task = args.pop('task', None)
        if task == 'clean':
            q = Query(mode='local')
            q._cleanUpQueries()
            res = {'status': 1, 'data': 'clean success'}
        elif task == 'getprocs':
            q = Query(mode='local')
            procs = q._getIdleProcesses()
            procs = [{k: v for k, v in y.items()} for y in procs]
            res = {'status': 1, 'data': procs}
        else:
            res = {'status': -1, 'data': None, 'error': 'Task is None or not in [clean, getprocs]'}
        self.update_results(res)
        output = jsonify(self.results)
        return output



#!/usr/bin/env python
# encoding: utf-8
#
# modelcube.py
#
# Licensed under a 3-clause BSD license.
#
# Revision history:
#     25 Sep 2016 J. Sánchez-Gallego
#       Initial version


from __future__ import division
from __future__ import print_function

from flask_classy import route
from flask import jsonify

from marvin.tools.modelcube import ModelCube
from marvin.api.base import BaseView, arg_validate as av
from marvin.core.exceptions import MarvinError
from marvin.utils.general import parseIdentifier


def _get_model_cube(name, **kwargs):
    """Retrieves a Marvin ModelCube object."""

    model_cube = None
    results = {}

    # Pop the release to remove a duplicate input to Maps
    release = kwargs.pop('release', None)

    # parse name into either mangaid or plateifu
    try:
        idtype = parseIdentifier(name)
    except Exception as err:
        results['error'] = 'Failed to parse input name {0}: {1}'.format(name, str(err))
        return model_cube, results

    try:
        if idtype == 'plateifu':
            plateifu = name
            mangaid = None
        elif idtype == 'mangaid':
            mangaid = name
            plateifu = None
        else:
            raise MarvinError('invalid plateifu or mangaid: {0}'.format(idtype))

        model_cube = ModelCube(mangaid=mangaid, plateifu=plateifu,
                               release=release, **kwargs)
        results['status'] = 1
    except Exception as err:
        results['error'] = 'Failed to retrieve ModelCube {0}: {1}'.format(name, str(err))

    return model_cube, results


class ModelCubeView(BaseView):
    """Class describing API calls related to ModelCubes."""

    route_base = '/modelcubes/'

    @route('/<name>/', defaults={'bintype': None, 'template_kin': None},
           methods=['GET', 'POST'], endpoint='getModelCube')
    @route('/<name>/<bintype>/', defaults={'template_kin': None},
           methods=['GET', 'POST'], endpoint='getModelCube')
    @route('/<name>/<bintype>/<template_kin>/', methods=['GET', 'POST'], endpoint='getModelCube')
    @av.check_args()
    def get(self, args, name, bintype, template_kin):
        """Retrieves a ModelCube.

        .. :quickref: ModelCube; Get a modelcube given a name, bintype, and template_kin

        :param name: The name of the maps as plate-ifu or mangaid
        :param bintype: The bintype associated with this maps.  If not defined, the default is used
        :param template_kin: The template_kin associated with this maps.  If not defined, the default is used
        :form release: the release of MaNGA data
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :json string header: the modelcube header as a string
        :json list shape: the modelcube shape [x, y]
        :json list wavelength: the modelcube wavelength array
        :json list redcorr: the modelcube redcorr array
        :json string wcs_header: the modelcube wcs_header as a string
        :json string bintype: the bintype of the modelcube
        :json string template_kin: the template library of the modelcube
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin2/api/modelcubes/8485-1901/SPX/GAU-MILESHC/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {"plateifu": 8485-1901,
                       "mangaid": '1-209232',
                       "header": "XTENSION= 'IMAGE', NAXIS=3, .... END",
                       "wcs_header": "WCSAXES = 3 / Number of coordindate axes .... END",
                       "shape": [34, 34],
                       "wavelength": [3621.6, 3622.43, 3623.26, ...],
                       "redcorr": [1.06588, 1.065866, 1.06585, ...],
                       "bintype": "SPX",
                       "template_kin": "GAU-MILESHC"
                      }
           }

        """

        # Pop any args we don't want going into ModelCube
        args = self._pop_args(args, arglist='name')

        model_cube, results = _get_model_cube(name, **args)

        self.update_results(results)

        if model_cube is not None:
            self.results['data'] = {
                'plateifu': model_cube.plateifu,
                'mangaid': model_cube.mangaid,
                'header': model_cube.header.tostring(),
                'shape': model_cube.shape,
                'wavelength': model_cube.wavelength.tolist(),
                'redcorr': model_cube.redcorr.tolist(),
                'wcs_header': model_cube.wcs.to_header_string(),
                'bintype': model_cube.bintype,
                'template_kin': model_cube.template_kin}

        return jsonify(self.results)



from datetime import datetime, time

import sqlalchemy as sa
from sqlalchemy_utils import EmailType
from wtforms.validators import (
    DataRequired,
    Email,
    InputRequired,
    Length,
    NumberRange,
    Optional
)
from wtforms_components import DateRange, TimeRange

from tests import ModelFormTestCase
from wtforms_alchemy import ClassMap, ModelForm, Unique


class TestAutoAssignedValidators(ModelFormTestCase):
    def test_auto_assigns_length_validators(self):
        self.init()
        self.assert_max_length('test_column', 255)

    def test_assigns_validators_from_info_field(self):
        self.init(info={'validators': Email()})
        self.assert_has_validator('test_column', Email)

    def test_assigns_unique_validator_for_unique_fields(self):
        self.init(unique=True)
        self.assert_has_validator('test_column', Unique)

    def test_assigns_non_nullable_fields_as_required(self):
        self.init(nullable=False)
        self.assert_has_validator('test_column', DataRequired)
        self.assert_has_validator('test_column', InputRequired)

    def test_type_level_not_nullable_validators(self):
        class ModelTest(self.base):
            __tablename__ = 'model_test'
            id = sa.Column(sa.Integer, primary_key=True)
            test_column = sa.Column(sa.Unicode(255), nullable=False)

        validator = DataRequired()

        class ModelTestForm(ModelForm):
            class Meta:
                model = ModelTest
                not_null_validator_type_map = ClassMap()
                not_null_validator = validator

        form = ModelTestForm()
        assert validator in form.test_column.validators

    def test_not_nullable_validator_with_type_decorator(self):
        class ModelTest(self.base):
            __tablename__ = 'model_test'
            id = sa.Column(sa.Integer, primary_key=True)
            test_column = sa.Column(EmailType, nullable=False)

        validator = DataRequired()

        class ModelTestForm(ModelForm):
            class Meta:
                model = ModelTest
                not_null_validator_type_map = ClassMap(
                    [(sa.String, validator)]
                )
                not_null_validator = []

        form = ModelTestForm()
        assert validator in form.test_column.validators

    def test_not_null_validator_as_empty_list(self):
        class ModelTest(self.base):
            __tablename__ = 'model_test'
            id = sa.Column(sa.Integer, primary_key=True)
            test_column = sa.Column(sa.Boolean, nullable=False)

        class ModelTestForm(ModelForm):
            class Meta:
                model = ModelTest
                not_null_validator_type_map = ClassMap()
                not_null_validator = []

        form = ModelTestForm()
        assert form.test_column.validators == []

    def test_not_null_validator_as_none(self):
        class ModelTest(self.base):
            __tablename__ = 'model_test'
            id = sa.Column(sa.Integer, primary_key=True)
            test_column = sa.Column(sa.Boolean, nullable=False)

        class ModelTestForm(ModelForm):
            class Meta:
                model = ModelTest
                not_null_validator_type_map = ClassMap()
                not_null_validator = None

        form = ModelTestForm()
        assert len(form.test_column.validators) == 1
        assert isinstance(form.test_column.validators[0], Optional)

    def test_not_nullable_booleans_are_required(self):
        self.init(sa.Boolean, nullable=False)
        self.assert_has_validator('test_column', InputRequired)

    def test_not_nullable_fields_with_defaults_are_not_required(self):
        self.init(nullable=False, default=u'default')
        self.assert_not_required('test_column')

    def test_assigns_nullable_integers_as_optional(self):
        self.init(sa.Integer, nullable=True)
        self.assert_optional('test_column')

    def test_override_email_validator(self):
        class ModelTest(self.base):
            __tablename__ = 'model_test'
            id = sa.Column(sa.Integer, primary_key=True)
            test_column = sa.Column(EmailType, nullable=True)

        def validator():
            return Email('Wrong email')

        class ModelTestForm(ModelForm):
            class Meta:
                model = ModelTest
                email_validator = validator

        form = ModelTestForm()
        assert form.test_column.validators[1].message == 'Wrong email'

    def test_override_optional_validator(self):
        class ModelTest(self.base):
            __tablename__ = 'model_test'
            id = sa.Column(sa.Integer, primary_key=True)
            test_column = sa.Column(EmailType, nullable=True)

        class MyOptionalValidator(object):
            def __init__(self, *args, **kwargs):
                pass

        class ModelTestForm(ModelForm):
            class Meta:
                model = ModelTest
                optional_validator = MyOptionalValidator

        form = ModelTestForm()
        assert isinstance(form.test_column.validators[0], MyOptionalValidator)

    def test_override_number_range_validator(self):
        class ModelTest(self.base):
            __tablename__ = 'model_test'
            id = sa.Column(sa.Integer, primary_key=True)
            test_column = sa.Column(sa.Integer, info={'min': 3}, nullable=True)

        def number_range(min=-1, max=-1):
            return NumberRange(min=min, max=max, message='Wrong number range')

        class ModelTestForm(ModelForm):
            class Meta:
                model = ModelTest
                number_range_validator = number_range

        form = ModelTestForm()
        assert form.test_column.validators[1].message == 'Wrong number range'

    def test_override_date_range_validator(self):
        class ModelTest(self.base):
            __tablename__ = 'model_test'
            id = sa.Column(sa.Integer, primary_key=True)
            test_column = sa.Column(
                sa.DateTime,
                info={'min': datetime(2000, 1, 1)},
                nullable=True
            )

        def date_range(min=None, max=None):
            return DateRange(min=min, max=max, message='Wrong date range')

        class ModelTestForm(ModelForm):
            class Meta:
                model = ModelTest
                date_range_validator = date_range

        form = ModelTestForm()
        assert form.test_column.validators[1].message == 'Wrong date range'

    def test_override_time_range_validator(self):
        class ModelTest(self.base):
            __tablename__ = 'model_test'
            id = sa.Column(sa.Integer, primary_key=True)
            test_column = sa.Column(
                sa.Time,
                info={'min': time(14, 30)},
                nullable=True
            )

        def time_range(min=None, max=None):
            return TimeRange(min=min, max=max, message='Wrong time')

        class ModelTestForm(ModelForm):
            class Meta:
                model = ModelTest
                time_range_validator = time_range

        form = ModelTestForm()
        assert form.test_column.validators[1].message == 'Wrong time'

    def test_override_length_validator(self):
        class ModelTest(self.base):
            __tablename__ = 'model_test'
            id = sa.Column(sa.Integer, primary_key=True)
            test_column = sa.Column(sa.Unicode(255), nullable=True)

        def length(min=-1, max=-1):
            return Length(min=min, max=max, message='Wrong length')

        class ModelTestForm(ModelForm):
            class Meta:
                model = ModelTest
                length_validator = length

        form = ModelTestForm()
        assert form.test_column.validators[1].message == 'Wrong length'

    def test_override_optional_validator_as_none(self):
        class ModelTest(self.base):
            __tablename__ = 'model_test'
            id = sa.Column(sa.Integer, primary_key=True)
            test_column = sa.Column(sa.Boolean, nullable=True)

        class ModelTestForm(ModelForm):
            class Meta:
                model = ModelTest
                optional_validator = None

        form = ModelTestForm()
        assert form.test_column.validators == []

    def test_override_unique_validator(self):
        class ModelTest(self.base):
            __tablename__ = 'model_test'
            id = sa.Column(sa.Integer, primary_key=True)
            test_column = sa.Column(
                sa.Unicode(255), unique=True, nullable=True
            )

        def unique(column, get_session):
            return Unique(
                column, get_session=get_session, message='Not unique'
            )

        class ModelTestForm(ModelForm):
            class Meta:
                model = ModelTest
                unique_validator = unique

            @staticmethod
            def get_session():
                return None

        form = ModelTestForm()
        assert form.test_column.validators[2].message == 'Not unique'



from __future__ import absolute_import, division, print_function, unicode_literals
# The line above will help with 2to3 support.

from os.path import isfile, exists, dirname, join
from re import search
from sdss_access import SDSSPath, AccessError
from sdss_access.sync.auth import Auth
from sdss_access.sync.stream import Stream


class RsyncAccess(SDSSPath):

    def __init__(self, label='sdss_rsync', stream_count=5, mirror=False, public=False, verbose=False):
        super(RsyncAccess, self).__init__(mirror=mirror, public=public, verbose=verbose)
        self.label = label
        self.auth = None
        self.stream = None
        self.stream_count = stream_count
        self.verbose = verbose
        self.initial_stream = self.get_stream()

    def get_stream(self):
        stream = Stream(stream_count=self.stream_count, verbose=self.verbose)
        return stream

    def remote(self, username=None, password=None):
        self.set_netloc(sdss=True)  # simplifies things to have a single sdss machine in .netrc
        self.set_auth(username=username, password=password)
        self.set_netloc(dtn=not self.public)
        self.set_remote_base(scheme="rsync")

    def set_auth(self, username=None, password=None):
        self.auth = Auth(public=self.public, netloc=self.netloc, verbose=self.verbose)
        self.auth.set_username(username)
        self.auth.set_password(password)
        if not self.public:
            if not self.auth.ready():
                self.auth.load()
            if not self.auth.ready():
                self.auth.set_username(inquire=True)
                self.auth.set_password(inquire=True)

    def reset(self):
        if self.stream:
            self.stream.reset()

    def add(self, filetype, **kwargs):
        location = self.location(filetype, **kwargs)
        source = self.url(filetype, sasdir='sas' if not self.public else '', **kwargs)
        destination = self.full(filetype, **kwargs)
        if location and source and destination:
            self.initial_stream.append_task(location=location, source=source, destination=destination)
        else:
            print("There is no file with filetype=%r to access in the tree module loaded" % filetype)

    def set_stream(self):
        if not self.auth:
            raise AccessError("Please use the remote() method to set rsync authorization or use remote(public=True) for public data")
        else:
            self.stream = self.get_stream()
            self.stream.source = join(self.remote_base, 'sas') if self.remote_base and not self.public else self.remote_base
            self.stream.destination = self.base_dir
            self.stream.cli.env = {'RSYNC_PASSWORD': self.auth.password} if self.auth.ready() else None
            if self.stream.source and self.stream.destination:
                for task in self.initial_stream.task:
                    self.set_stream_task(task)

    def get_task_out(self, task=None):
        if task:
            command = "rsync -R %(source)s" % task
            if self.verbose:
                print("rsync -R %(source)s" % task)
            status, out, err = self.stream.cli.foreground_run(command)
            if status:
                raise AccessError("Return code %r\n%s" % (status, err))
        else:
            out = None
        return out

    def generate_stream_task(self, task=None, out=None):
        if task and out:
            release = task['location'].split('/')[0]
            depth = task['location'].count('/')
            for result in out.split(b"\n"):
                result = result.decode('utf-8')
                if result.startswith(('d', '-', 'l')):
                    try:
                        location = search(r"^.*\s{1,3}(.+)$", result).group(1)
                    except:
                        location = None
                    if self.public:
                        location = join(release, location)
                    if location and location.count('/') == depth:
                        source = join(self.stream.source, location) if self.remote_base else None
                        destination = join(self.stream.destination, location)
                        yield (location, source, destination)

    def set_stream_task(self, task=None):
        out = self.get_task_out(task=task)
        stream_has_task = False
        for location, source, destination in self.generate_stream_task(task=task, out=out):
            if location and source and destination:
                stream_has_task = True
                self.stream.append_task(location=location, source=source, destination=destination)
                """if self.verbose:
                    print("SDSS_ACCESS> Preparing to download: %s" % location)
                    print("SDSS_ACCESS> from: %s" % source)
                    print("SDSS_ACCESS> to: %s" % destination)
                    print("-"*80)"""

        if not stream_has_task:
            print('SDSS_ACCESS> Error: stream has nothing to do.')

    def shuffle(self):
        self.stream.shuffle()

    def get_locations(self, offset=None, limit=None):
        return self.stream.get_locations(offset=offset, limit=limit) if self.stream else None

    def get_paths(self, offset=None, limit=None):
        locations = self.get_locations(offset=offset, limit=limit)
        paths = [join(self.base_dir, location) for location in locations] if locations else None
        return paths

    def get_urls(self, offset=None, limit=None):
        locations = self.get_locations(offset=offset, limit=limit)
        remote_base = self.get_remote_base()
        sasdir = 'sas' if not self.public else ''
        urls = [join(remote_base, sasdir, location) for location in locations] if locations else None
        return urls

    def refine_task(self, regex=None):
        self.stream.refine_task(regex=regex)

    def commit(self, offset=None, limit=None, dryrun=False):
        self.stream.command = "rsync -avRK --files-from={path} {source} {destination}"
        self.stream.append_tasks_to_streamlets(offset=offset, limit=limit)
        self.stream.commit_streamlets()
        self.stream.run_streamlets()
        self.stream.reset_streamlet()



#!/usr/bin/env python
# encoding: utf-8
#
# file.py
#
# Licensed under a 3-clause BSD license.
#
# Revision history:
#    21 Apr 2016 J. Sánchez-Gallego
#       Initial version


from __future__ import division
from __future__ import print_function
from brain.core.exceptions import BrainError


class URLMapDict(dict):
    """A custom dictionary for urlmap that fails with a custom error."""

    def __init__(self, inp={}):
        self.update(**dict((kk, self.parse(vv)) for kk, vv in inp.items()))

    @classmethod
    def parse(cls, vv):
        if isinstance(vv, dict):
            return cls(vv)
        elif isinstance(vv, list):
            return [cls.parse(ii) for ii in vv]
        else:
            return vv

    def __missing__(self, key):
        """Overrides the default KeyError exception."""

        if len(self.keys()) == 0:
            raise BrainError('No URL Map found. Cannot make remote call')
        else:
            raise BrainError('Key {0} not found in urlmap.'.format(key))



#!/usr/bin/env python
# encoding: utf-8
#
# test_bin.py
#
# Created by José Sánchez-Gallego on 6 Nov 2016.
# Modified by Brett Andrews on 24 Apr 2017.


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os

import pytest

import marvin
import marvin.tools.maps
import marvin.tools.modelcube
import marvin.utils.general
import marvin.tests
from marvin.tools.bin import Bin
from marvin.tests import UseBintypes, marvin_test_if_class
from marvin.core.exceptions import MarvinError


@UseBintypes('VOR10')
@pytest.mark.xfail(run=False, reason='expected fail until someone fixes it')
class TestBinInit:

    def _check_bin_data(self, bb, gal):

        assert bb.binid == 100
        assert bb.plateifu == gal.plateifu
        assert bb.mangaid == gal.mangaid

        assert len(bb.spaxels) == 2
        assert not bb.spaxels[0].loaded

        assert bb.properties is not None

    def test_init_from_files(self, galaxy):

        bb = Bin(binid=100, maps_filename=galaxy.mapspath, modelcube_filename=galaxy.modelpath)

        assert isinstance(bb._maps, marvin.tools.maps.Maps)
        assert isinstance(bb._modelcube, marvin.tools.modelcube.ModelCube)

        self._check_bin_data(bb, galaxy)

    def test_init_from_file_only_maps(self, galaxy):

        bb = Bin(binid=100, maps_filename=galaxy.mapspath)

        assert isinstance(bb._maps, marvin.tools.maps.Maps)
        assert bb._modelcube is not None
        assert bb._modelcube.data_origin == 'db'
        assert bb._modelcube.bintype == galaxy.bintype

        self._check_bin_data(bb, galaxy)

    def test_init_from_db(self, galaxy):

        bb = Bin(binid=100, plateifu=galaxy.plateifu, bintype=galaxy.bintype)
        assert bb._maps.data_origin == 'db'
        assert isinstance(bb._maps, marvin.tools.maps.Maps)
        assert isinstance(bb._modelcube, marvin.tools.modelcube.ModelCube)
        assert bb._modelcube.bintype == galaxy.bintype

        self._check_bin_data(bb, galaxy)

    def test_init_from_api(self, galaxy):

        bb = Bin(binid=100, plateifu=galaxy.plateifu, mode='remote', bintype=galaxy.bintype)

        assert isinstance(bb._maps, marvin.tools.maps.Maps)
        assert isinstance(bb._modelcube, marvin.tools.modelcube.ModelCube)
        assert bb._modelcube.bintype == galaxy.bintype

        self._check_bin_data(bb, galaxy)

    def test_bin_does_not_exist(self, galaxy):

        with pytest.raises(MarvinError) as ee:
            Bin(binid=99999, plateifu=galaxy.plateifu, mode='local', bintype=galaxy.bintype)
            assert 'there are no spaxels associated with binid=99999.' in str(ee.exception)


#@marvin_test_if_class(mark='skip', galaxy=dict(release=['MPL-4']))
@pytest.mark.xfail(run=False, reason='expected fail until someone fixes it')
class TestBinFileMismatch:
    def test_bintypes(self, galaxy):

        wrong_bintype = 'WRONGSPX'
        assert wrong_bintype != galaxy.bintype

        wrong_modelcube_filename = galaxy.new_path('modelcube', {'bintype': wrong_bintype, 'daptype': '{0}-GAU-MILESHC'.format(wrong_bintype)})

        with pytest.raises(MarvinError) as cm:
            bb = Bin(binid=100, maps_filename=galaxy.mapspath, modelcube_filename=wrong_modelcube_filename)

        assert 'there are no spaxels associated with binid=100.' in str(cm.value)

        # assert isinstance(bb._maps, marvin.tools.maps.Maps)
        # assert isinstance(bb._modelcube, marvin.tools.modelcube.ModelCube)

        # with pytest.raises(MarvinError):
        #     marvin.utils.general._check_file_parameters(bb._maps, bb._modelcube)()



from __future__ import absolute_import, division, print_function, unicode_literals
# The line above will help with 2to3 support.

from sdss_access.sync import Cli
from random import shuffle
from re import compile

class Stream:

    max_stream_count = 5

    def __init__(self, stream_count=None, verbose=False):
        self.verbose = verbose
        try: self.stream_count = min(int(stream_count),self.max_stream_count)
        except: self.stream_count = 0
        self.streamlet = [{'index':index, 'location':None, 'source':None, 'destination':None}  for index in range(0,self.stream_count)]
        self.reset()
        self.index = 0
        self.command = None
        self.env = None
        self.source = None
        self.destination = None
        self.cli = Cli(verbose=verbose)

    def reset(self):
        self.reset_task()
        self.reset_streamlet()
        
    def reset_task(self): self.task = []

    def reset_streamlet(self):
        for index in range(0,self.stream_count): self.set_streamlet(index=index, location=[], source=[], destination=[])

    def set_streamlet(self, index=None, location=None, source=None, destination=None):
        streamlet = self.get_streamlet(index=index)
        if streamlet:
            try:
                n = len(location)
                ok = n == len(source) and n == len(destination)
                streamlet['location'], streamlet['source'], streamlet['destination'] = (location, source, destination) if ok else (None, None, None)
            except:
                streamlet['location'], streamlet['source'], streamlet['destination'] = (None, None, None)


    def get_streamlet(self,index=None,increment=1):
        if index is None:
            self.index += increment
            if self.index >= self.stream_count: self.index = 0
        else:
            try: self.index = int(index)
            except: self.index = 0
        try: streamlet = self.streamlet[self.index]
        except: streamlet = None
        return streamlet

    def get_locations(self, offset=None, limit=None):
        locations = [task['location'] for task in self.task] if self.task else None
        if offset: locations = locations[offset:]
        if limit: locations = locations[:limit]
        return locations
    
    def shuffle(self): shuffle(self.task)
    
    def refine_task(self, regex=None):
        locations = self.get_locations()
        r = compile(regex)
        subset = filter(lambda i: r.search(i), locations)
        self.task = [self.task[locations.index(s)] for s in subset]

    def append_task(self, location=None, source=None, destination=None):
        if location and source and destination:
            task = {'location':location, 'source':source, 'destination':destination, 'exists':None}
            self.task.append(task)

    def append_tasks_to_streamlets(self, offset=None, limit=None):
        tasks = []
        ntasks = 0
        for index,task in enumerate(self.task):
            if (offset is None or index>=offset):
                tasks.append(task)
                ntasks += 1
            if limit is not None and ntasks >= limit: break
        for task in tasks: self.append_streamlet(task=task)
    
    def append_streamlet(self, index=None, task=None):
        streamlet = self.get_streamlet(index=index)
        if streamlet and task:
                streamlet['location'].append(task['location'])
                streamlet['source'].append(task['source'])
                streamlet['destination'].append(task['destination'])

    def commit_streamlets(self):
        if self.command:
            self.cli.set_dir()
            for streamlet in self.streamlet: self.commit_streamlet(streamlet)
            if self.verbose: print("SDSS_ACCESS> streamlets added to {dir}".format(dir=self.cli.dir))

    def commit_streamlet(self, streamlet=None):
        if streamlet:
            streamlet['path'] = self.cli.get_path(index=streamlet['index'])
            path_txt = "{0}.txt".format(streamlet['path'])
            streamlet['command'] = self.command.format(path=path_txt,source=self.source,destination=self.destination)
            self.cli.write_lines(path=path_txt, lines=[location for location in streamlet['location']])

    def run_streamlets(self):
        for streamlet in self.streamlet:
            streamlet['logfile'] = open("{0}.log".format(streamlet['path']),"w")
            streamlet['errfile'] = open("{0}.err".format(streamlet['path']),"w")
            streamlet['process'] = self.cli.get_background_process(streamlet['command'], logfile=streamlet['logfile'], errfile=streamlet['errfile'])
            if self.verbose: print("SDSS_ACCESS> rsync stream %s logging to %s" % (streamlet['index'],streamlet['logfile'].name))
        self.cli.wait_for_processes(streamlet['process'] for streamlet in self.streamlet)
        if self.cli.returncode: print("SDSS_ACCESS> return code {returncode}".format(returncode=self.cli.returncode))
        for streamlet in self.streamlet: streamlet['logfile'].close()



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-05-25 10:11:21
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-07-14 17:13:58

from __future__ import print_function, division, absolute_import
from marvin.tools.query import Query
from marvin.core.exceptions import MarvinError
from marvin import config
from marvin.tools.cube import Cube
from marvin.tools.maps import Maps
from marvin.tools.spaxel import Spaxel
from marvin.tools.modelcube import ModelCube
import pytest


class TestQueryVersions(object):

    def test_versions(self, query, release, versions):
        drpver, dapver = versions
        assert query._release == release
        assert query._drpver == drpver
        assert query._dapver == dapver


class TestQuerySearches(object):

    @pytest.mark.parametrize('query, joins',
                             [('nsa.z < 0.1', ['nsa', 'drpalias']),
                              ('haflux > 25', ['spaxelprop', 'dapalias']),
                              ('nsa.z < 0.1 and haflux > 25', ['nsa', 'spaxelprop', 'drpalias', 'dapalias'])],
                             ids=['drponly', 'daponly', 'drp_and_dap'],
                             indirect=['query'])
    def test_whereclause(self, query, joins):
        if query.mode == 'remote':
            res = query.run()
            for item in joins:
                assert item in str(res.query)
        else:
            for item in joins:
                assert item in str(query.query.whereclause)

    @pytest.mark.parametrize('query, addparam',
                             [('nsa.z < 0.1', ['nsa.z']),
                              ('haflux > 25', ['emline_gflux_ha_6564', 'spaxelprop.x', 'spaxelprop.y', 'bintype.name', 'template.name'])],
                             indirect=['query'])
    def test_params(self, query, addparam):
        params = query.expdata['defaults'] + addparam
        res = query.run()
        assert set(params) == set(query.params)

    @pytest.mark.parametrize('badquery, errmsg',
                             [('nsa.hello < 0.1', 'nsa.hello does not match any column.'),
                              ('name = SPX', 'name matches multiple parameters in the lookup table'),
                              ('< 0.1', 'Your boolean expression contained a syntax error')],
                             ids=['nomatch', 'multiple_entries', 'syntax_error'])
    def test_bad_queries(self, expmode, badquery, errmsg):
        if expmode is None:
            pytest.skip('cannot use queries in local mode without a db')

        with pytest.raises(MarvinError) as cm:
            query = Query(searchfilter=badquery, mode=expmode)
            res = query.run()
        assert cm.type == MarvinError
        assert errmsg in str(cm.value)
        if expmode != 'local':
            assert config._traceback is not None

    @pytest.mark.parametrize('query, allspax, table',
                             [('haflux > 25', False, 'cleanspaxelprop'),
                              ('haflux > 25', True, 'spaxelprop')],
                             ids=['allspax', 'cleanspax'],
                             indirect=['query'])
    def test_spaxel_tables(self, query, expmode, allspax, table):
        table = table + config.release.split('-')[1] if '4' not in config.release else table
        query = Query(searchfilter=query.searchfilter, allspaxels=allspax, mode=query.mode, release=query._release)
        if expmode == 'local':
            assert table in set(query.joins)
        else:
            res = query.run()
            assert table in res.query

    @pytest.mark.parametrize('query, sfilter',
                             [('nsa.z < 0.1', 'nsa.z < 0.1'),
                              ('abs_g_r > -1', 'abs_g_r > -1'),
                              ('haflux > 25', 'haflux > 25'),
                              ('npergood(emline_gflux_ha_6564 > 5) > 20', 'npergood(emline_gflux_ha_6564 > 5) > 20'),
                              ('nsa.z < 0.1 and haflux > 25', 'nsa.z < 0.1 and haflux > 25')],
                             indirect=['query'], ids=['nsaz', 'absgr', 'haflux', 'npergood', 'nsahaflux'])
    def test_success_queries(self, query, sfilter):
        res = query.run()
        count = query.expdata['queries'][sfilter]
        assert count['count'] == res.totalcount


class TestQueryReturnParams(object):

    @pytest.mark.parametrize('query', [('nsa.z < 0.1')], indirect=True)
    @pytest.mark.parametrize('rps', [(['g_r']), (['cube.ra', 'cube.dec']), (['haflux'])])
    def test_success(self, query, rps):
        query = Query(searchfilter=query.searchfilter, returnparams=rps, mode=query.mode)
        assert 'nsa.z' in query.params
        assert set(rps).issubset(set(query.params))

    @pytest.mark.parametrize('query', [('nsa.z < 0.1')], indirect=True)
    @pytest.mark.parametrize('rps, errmsg',
                             [('hello', 'does not match any column.'),
                              ('name', 'name matches multiple parameters')],
                             ids=['nomatch', 'multiple_entries'])
    def test_badparams(self, query, expmode, rps, errmsg):
        # set error type based on query mode
        if expmode == 'remote':
            error = MarvinError
        else:
            error = KeyError

        with pytest.raises(error) as cm:
            query = Query(searchfilter=query.searchfilter, returnparams=[rps], mode=query.mode)
            res = query.run()
        assert cm.type == error
        assert errmsg in str(cm.value)


class TestQueryReturnType(object):

    @pytest.mark.parametrize('query', [('cube.mangaid == 1-209232 and haflux > 25')], indirect=True)
    @pytest.mark.parametrize('objtype, tool',
                             [('cube', Cube), ('maps', Maps), ('spaxel', Spaxel),
                              ('modelcube', ModelCube)])
    def test_get_success(self, query, objtype, tool):
        if query.mode == 'remote' and config.db is None:
            pytest.skip('skipping weird case where nodb, remote mode tried to load a local file')
        if config.release == 'MPL-4' and objtype == 'modelcube':
            pytest.skip('no modelcubes in mpl-4')

        query = Query(searchfilter=query.searchfilter, returntype=objtype, mode=query.mode, release=query._release)
        res = query.run()
        assert res.objects is not None
        assert len(res.results) == len(res.objects)
        assert isinstance(res.objects[0], tool) is True

    @pytest.mark.parametrize('query', [('nsa.z < 0.1')], indirect=True)
    @pytest.mark.parametrize('objtype, errmsg',
                             [('noncube', 'Query returntype must be either cube, spaxel, maps, modelcube, rss')])
    def test_badreturntype(self, query, objtype, errmsg):
        with pytest.raises(AssertionError) as cm:
            query = Query(searchfilter=query.searchfilter, returntype=objtype, mode=query.mode)
        assert cm.type == AssertionError
        assert errmsg in str(cm.value)


class TestQueryPickling(object):

    @pytest.mark.parametrize('query', [('nsa.z < 0.1')], indirect=True)
    def test_pickle_save(self, temp_scratch, query):
        if query.mode == 'local':
            pytest.xfail('save cannot be run in local mode')
        file = temp_scratch.join('test_query.mpf')
        query.save(str(file))
        assert file.check() is True

    @pytest.mark.parametrize('query, sfilter', [('nsa.z < 0.1', 'nsa.z < 0.1')], indirect=['query'])
    def test_pickle_restore(self, temp_scratch, query, sfilter):
        if query.mode == 'local':
            pytest.xfail('save cannot be run in local mode')
        file = temp_scratch.join('test_query.mpf')
        query.save(str(file))
        assert file.check() is True
        query = None
        assert query is None
        query = Query.restore(str(file))
        assert query.searchfilter == sfilter


class TestQueryParams(object):

    @pytest.mark.parametrize('paramdisplay', [('all'), ('best')])
    def test_getparams(self, query, paramdisplay):
        params = query.get_available_params(paramdisplay)
        mydata = query.expdata['params'][paramdisplay]
        # counts and content
        if paramdisplay == 'best':
            assert mydata['count'] == sum([len(v) for v in params])
            assert set(mydata['subset']).issubset(set(params.list_params()))
            assert set(mydata).isdisjoint(set(params.list_params()))
        elif paramdisplay == 'all':
            assert mydata['count'] == len(params)
            assert set(mydata['subset']).issubset(set(params))


class TestQueryModes(object):

    @pytest.mark.parametrize('query', [('nsa.z < 0.1 and cube.plate == 8485')], indirect=True)
    def test_getmode(self, query, expmode):
        assert query.mode == expmode
        res = query.run()
        assert res.mode == expmode
        assert query.mode == res.mode



import sqlalchemy as sa
from pytest import mark
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy_utils.types import phone_number

from tests import MultiDict
from wtforms_alchemy import ModelForm


class TestCase(object):
    def setup_method(self, method):
        self.engine = create_engine('sqlite:///:memory:')
        self.Base = declarative_base()

        self.create_models()
        self.Base.metadata.create_all(self.engine)

        Session = sessionmaker(bind=self.engine)
        self.session = Session()

    def teardown_method(self, method):
        self.session.close_all()
        self.Base.metadata.drop_all(self.engine)
        self.engine.dispose()

    def create_models(self):
        class User(self.Base):
            __tablename__ = 'user'
            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
            name = sa.Column(sa.Unicode(255))
            phone_number = sa.Column(
                phone_number.PhoneNumberType(country_code='FI')
            )

        self.User = User


@mark.xfail('phone_number.phonenumbers is None')
class TestPhoneNumbers(TestCase):
    '''
    Simple tests to ensure that sqlalchemy_utils.PhoneNumber,
    wtforms_alchemy.PhoneNumberType and sqlalchemy_utils.PhoneNumberField work
    nicely together.
    '''
    def setup_method(self, method):
        super(TestPhoneNumbers, self).setup_method(method)

        class UserForm(ModelForm):
            class Meta:
                model = self.User
        self.UserForm = UserForm

        super(TestPhoneNumbers, self).setup_method(method)
        self.phone_number = phone_number.PhoneNumber(
            '040 1234567',
            'FI'
        )
        self.user = self.User()
        self.user.name = u'Someone'
        self.user.phone_number = self.phone_number
        self.session.add(self.user)
        self.session.commit()

    def test_query_returns_phone_number_object(self):
        queried_user = self.session.query(self.User).first()
        assert queried_user.phone_number == self.phone_number

    def test_phone_number_is_stored_as_string(self):
        result = self.session.execute(
            'SELECT phone_number FROM user WHERE id=:param',
            {'param': self.user.id}
        )
        assert result.first()[0] == u'+358401234567'

    def test_phone_number_in_form(self):
        form = self.UserForm(MultiDict(
            name=u'Matti Meikalainen',
            phone_number='+358401231233'
        ))
        form.validate()
        assert len(form.errors) == 0
        assert form.data['phone_number'] == (
            phone_number.PhoneNumber('+358401231233')
        )

    def test_empty_phone_number_in_form(self):
        form = self.UserForm(MultiDict(
            name=u'Matti Meikalainen',
            phone_number=''
        ))
        form.validate()
        assert len(form.errors) == 0
        assert form.data['phone_number'] is None



#!/usr/bin/env python
# License information goes here
#
# Imports
#
import glob
import os
import sys
from setuptools import setup, find_packages
setup_keywords = dict()
#
# THESE SETTINGS NEED TO BE CHANGED FOR EVERY PRODUCT.
#
setup_keywords['name'] = 'sdss_access'
setup_keywords['description'] = 'sdss_access package'
setup_keywords['author'] = 'Michael Blanton'
setup_keywords['author_email'] = 'michael.blanton@nyu.edu'
setup_keywords['license'] = 'BSD-3'
setup_keywords['url'] = 'https://svn.sdss.org/repo/sdss/sdss_access'
#
# END OF SETTINGS THAT NEED TO BE CHANGED.
#
#
# Import this module to get __doc__ and __version__.
#
sys.path.insert(int(sys.path[0] == ''), './python')
try:
    from importlib import import_module
    product = import_module(setup_keywords['name'])
    setup_keywords['long_description'] = product.__doc__
    setup_keywords['version'] = product.__version__
except ImportError:
    #
    # Try to get the long description from the README.rst file.
    #
    if os.path.exists('README.rst'):
        with open('README.rst') as readme:
            setup_keywords['long_description'] = readme.read()
    else:
        setup_keywords['long_description'] = ''
    setup_keywords['version'] = '0.0.1.dev'
#
# Indicates if this version is a release version.
#
if setup_keywords['version'].endswith('dev'):
    #
    # Try to obtain svn information.
    #
    try:
        from sdss4tools.install import get_svn_devstr
        setup_keywords['version'] += get_svn_devstr()
    except ImportError:
        pass
#
# Set other keywords for the setup function.  These are automated, & should
# be left alone unless you are an expert.
#
# Treat everything in bin/ except *.rst as a script to be installed.
#
if os.path.isdir('bin'):
    setup_keywords['scripts'] = \
        [fname for fname in glob.glob(os.path.join('bin', '*'))
         if not os.path.basename(fname).endswith('.rst')]
setup_keywords['provides'] = [setup_keywords['name']]
setup_keywords['requires'] = ['Python (>2.6.0)']
# setup_keywords['install_requires'] = ['Python (>2.6.0)']
setup_keywords['zip_safe'] = False
setup_keywords['use_2to3'] = True
setup_keywords['packages'] = find_packages('py')
setup_keywords['package_dir'] = {'': 'py'}
#
# Run setup command.
#
setup(**setup_keywords)



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-06-12 19:13:30
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-06-12 19:23:07

from __future__ import print_function, division, absolute_import
from marvin.core import marvin_pickle
import pytest
import os

data = dict(a=7, b=[10, 2])


class TestMarvinPickle(object):

    def test_specify_path(self, temp_scratch):
        file = temp_scratch.join('tmp_testMarvinPickleSpecifyPath.pck')
        path_out = marvin_pickle.save(obj=data, path=str(file), overwrite=False)
        assert file.check() is True
        revived_data = marvin_pickle.restore(path_out)
        assert data == revived_data

    def test_overwrite_true(self, temp_scratch):
        file = temp_scratch.join('tmp_testMarvinPickleOverwriteTrue.pck')
        open(str(file), 'a').close()
        path_out = marvin_pickle.save(obj=data, path=str(file), overwrite=True)
        assert file.check() is True
        revived_data = marvin_pickle.restore(path_out)
        assert data == revived_data

    def test_delete_on_restore(self, temp_scratch):
        file = temp_scratch.join('tmp_testMarvinPickleDeleteOnRestore.pck')
        path_out = marvin_pickle.save(obj=data, path=str(file), overwrite=False)
        assert file.check() is True
        revived_data = marvin_pickle.restore(path_out, delete=True)
        assert data == revived_data
        assert os.path.isfile(str(file)) is False





# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-05-19 16:34:31
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-07-30 19:32:03

from __future__ import print_function, division, absolute_import
from marvin.tests.api.conftest import ApiPage
import pytest


@pytest.mark.parametrize('page', [('api', 'mangaid2plateifu')], ids=['mangaid2plateifu'], indirect=True)
class TestGeneralMangaid2Plateifu(object):

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    def test_getplateifu_success(self, galaxy, page, params, reqtype):
        data = galaxy.plateifu
        page.load_page(reqtype, page.url.format(mangaid=galaxy.mangaid), params=params)
        page.assert_success(data)

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('mangaid', [('1209232')], ids=['badid'])
    def test_getplateifu_noresult(self, mangaid, page, params, reqtype):
        data = None
        error = "manga2plateifu failed with error: no plate-ifus found for mangaid={0}".format(mangaid)
        page.load_page(reqtype, page.url.format(mangaid=mangaid), params=params)
        assert page.json['status'] == -1
        assert page.json['error'] == error

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('mangaid, missing, errmsg', [(None, 'release', 'Missing data for required field.'),
                                                          ('12', 'mangaid', 'Length must be between 4 and 20.')],
                             ids=['norelease', 'shortname'])
    def test_getplateifu_failure(self, galaxy, page, reqtype, params, mangaid, missing, errmsg):
        if mangaid is None:
            page.route_no_valid_params(page.url.format(mangaid=galaxy.mangaid), missing, reqtype=reqtype, errmsg=errmsg)
        else:
            page.route_no_valid_params(page.url.format(mangaid=mangaid), missing, reqtype=reqtype, params=params, errmsg=errmsg)


@pytest.mark.parametrize('page', [('api', 'nsa_full')], ids=['nsa_full'], indirect=True)
class TestGeneralNSAFull(object):

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    def test_getnsa_success(self, galaxy, page, params, reqtype):
        data = {'elpetro_absmag_i': -19.1125469207764, 'elpetro_mag_g_r': 0.64608402745868077, 'z': 0.0407447,
                'elpetro_th50_r': 1.33067, 'elpetro_logmass': 9.565475912843823,
                'elpetro_ba': 0.87454, 'elpetro_mag_i_z': 0.2311751372102151,
                'elpetro_phi': 154.873, 'elpetro_mtol_i': 1.30610692501068,
                'elpetro_th90_r': 3.6882, 'elpetro_mag_u_r': 1.8892372699482216,
                'sersic_n': 3.29617}
        page.load_page(reqtype, page.url.format(mangaid=galaxy.mangaid), params=params)
        page.assert_success(galaxy.nsa_data['nsa'])

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('mangaid', [('1209232')], ids=['badid'])
    def test_getnsa_noresult(self, mangaid, page, params, reqtype):
        error = "get_nsa_data failed with error: get_nsa_data: cannot find NSA row for mangaid={0}".format(mangaid)
        page.load_page(reqtype, page.url.format(mangaid=mangaid), params=params)
        assert page.json['data'] is None
        assert page.json['status'] == -1
        assert page.json['error'] == error


@pytest.mark.parametrize('page', [('api', 'nsa_drpall')], ids=['nsa_drpall'], indirect=True)
class TestGeneralNSADrpall(object):

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    def test_getnsa_success(self, galaxy, page, params, reqtype):
        data = {'elpetro_absmag_i': -19.1125469207764, 'elpetro_mag_g_r': 0.64608402745868077, 'z': 0.0407447,
                'elpetro_th50_r': 1.33067, 'elpetro_logmass': 9.565475912843823,
                'elpetro_ba': 0.87454, 'elpetro_mag_i_z': 0.2311751372102151,
                'elpetro_phi': 154.873, 'elpetro_mtol_i': 1.30610692501068,
                'elpetro_th90_r': 3.6882, 'elpetro_mag_u_r': 1.8892372699482216,
                'sersic_n': 3.29617}
        page.load_page(reqtype, page.url.format(mangaid=galaxy.mangaid), params=params)
        page.assert_success(galaxy.nsa_data['drpall'])

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('mangaid', [('1209232')], ids=['badid'])
    def test_getnsa_noresult(self, mangaid, page, params, reqtype):
        error = "get_nsa_data failed with error: no plate-ifus found for mangaid={0}".format(mangaid)
        page.load_page(reqtype, page.url.format(mangaid=mangaid), params=params)
        assert page.json['data'] is None
        assert page.json['status'] == -1
        assert page.json['error'] == error



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2016-12-08 14:24:58
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-06-28 17:53:20

from __future__ import print_function, division, absolute_import
from flask_classy import FlaskView
from flask import request
from marvin.web.web_utils import parseSession
import marvin
from brain.api.general import BrainGeneralRequestsView
from marvin.api.base import arg_validate as av
import json


class BaseWebView(FlaskView):
    ''' This is the Base Web View for all pages '''

    def __init__(self, page):
        self.base = {}
        self.base['intro'] = 'Welcome to Marvin!'
        self.base['version'] = marvin.__version__
        self.update_title(page)
        self._endpoint = self._release = None
        self._drpver = self._dapver = None

    def before_request(self, *args, **kwargs):
        ''' this runs before every single request '''
        self.base['error'] = None
        self._endpoint = request.endpoint
        self._drpver, self._dapver, self._release = parseSession()

        # try to get a local version of the urlmap for the arg_validator
        if not av.urlmap:
            bv = BrainGeneralRequestsView()
            resp = bv.buildRouteMap()
            marvin.config.urlmap = json.loads(resp.get_data())['urlmap']
            av.urlmap = marvin.config.urlmap

    def after_request(self, name, response):
        ''' this runs after every single request '''

        return response

    def update_title(self, page):
        ''' Update the title and page '''
        self.base['title'] = page.title().split('-')[0] if 'main' in page \
            else page.title().replace('-', ' | ')
        self.base['page'] = page

    def reset_dict(self, mydict, exclude=None):
        ''' resets the page dictionary '''
        mydict['error'] = self.base['error']
        exclude = exclude if isinstance(exclude, list) else [exclude]
        diffkeys = set(mydict) - set(self.base)
        for key, val in mydict.items():
            if key in diffkeys and (key not in exclude):
                mydict[key] = '' if isinstance(val, str) else None



from flask import current_app, Blueprint, render_template, jsonify
from flask import session as current_session, request, redirect, url_for
from flask_classy import route
from marvin import config, marvindb
from brain.api.base import processRequest
from marvin.utils.general.general import parseIdentifier
from marvin.web.web_utils import parseSession
from marvin.api.base import arg_validate as av
from marvin.web.controllers import BaseWebView

from hashlib import md5
# try:
#     print('importing main inspection')
#     from inspection.marvin import Inspection
# except ImportError as e:
#     print('importing local inspection')
#     from brain.core.inspection import Inspection

index = Blueprint("index_page", __name__)


class Marvin(BaseWebView):
    route_base = '/'

    def __init__(self):
        super(Marvin, self).__init__('marvin-main')
        self.main = self.base.copy()

    def before_request(self, *args, **kwargs):
        super(Marvin, self).before_request(*args, **kwargs)
        self.reset_dict(self.main)

    def index(self):
        current_app.logger.info('Welcome to Marvin Web!')

        return render_template("index.html", **self.main)

    def quote(self):
        return 'getting quote'

    def status(self):
        return 'OK'

    @route('/test/')
    def test(self):
        return 'new test'

    def database(self):
        onecube = marvindb.session.query(marvindb.datadb.Cube).order_by(marvindb.datadb.Cube.pk).first()
        return jsonify(result={'plate': onecube.plate, 'status': 1})

    @route('/galidselect/', methods=['GET', 'POST'], endpoint='galidselect')
    def galidselect(self):
        ''' Route that handle the Navbar plate/galaxy id search form '''
        args = av.manual_parse(self, request, use_params='index', required='galid')
        galid = args.get('galid', None)
        if not galid:
            # if not galid return main page
            return redirect(url_for('index_page.Marvin:index'))
        else:
            idtype = parseIdentifier(galid)
        # check the idtype
        if idtype == 'plateifu' or idtype == 'mangaid':
            return redirect(url_for('galaxy_page.Galaxy:get', galid=galid))
        elif idtype == 'plate':
            return redirect(url_for('plate_page.Plate:get', plateid=galid))
        else:
            return redirect(url_for('index_page.Marvin:index'))

    @route('/getgalidlist/', methods=['GET', 'POST'], endpoint='getgalidlist')
    def getgalidlist(self):
        ''' Retrieves the list of galaxy ids and plates for Bloodhound Typeahead '''

        if marvindb.datadb is None:
            out = ['', '', '']
            current_app.logger.info('ERROR: Problem with marvindb.datadb.  Cannot build galaxy id auto complete list.')
        else:
            cubes = (marvindb.session.query(marvindb.datadb.Cube.plate, marvindb.datadb.Cube.mangaid,
                                            marvindb.datadb.Cube.plateifu).join(marvindb.datadb.PipelineInfo,
                                                                                marvindb.datadb.PipelineVersion,
                                                                                marvindb.datadb.IFUDesign).
                     filter(marvindb.datadb.PipelineVersion.version == self._drpver).all())
            out = [str(e) for l in cubes for e in l]
        out = list(set(out))
        out.sort()
        return jsonify(out)

    @route('/selectmpl/', methods=['GET', 'POST'], endpoint='selectmpl')
    def selectmpl(self):
        ''' Global selection of the MPL/DR versions '''
        args = av.manual_parse(self, request, use_params='index')
        version = args.get('release', None)
        current_session['release'] = version
        drpver, dapver = config.lookUpVersions(release=version)
        current_session['drpver'] = drpver
        current_session['dapver'] = dapver
        out = {'status': 1, 'msg': 'Success', 'current_release': version,
               'current_drpver': drpver, 'current_dapver': dapver}

        return jsonify(result=out)

    @route('/login/', methods=['GET', 'POST'], endpoint='login')
    def login(self):
        ''' login for trac user '''
        try:
            from inspection.marvin import Inspection
        except ImportError as e:
            from brain.core.inspection import Inspection

        form = processRequest(request=request)
        result = {}
        username = form['username'].strip()
        password = form['password'].strip()
        auth = md5("{0}:AS3Trac:{1}".format(username, password).encode('utf-8')).hexdigest() if username and password else None
        try:
            inspection = Inspection(current_session, username=username, auth=auth)
        except Exception as e:
            result['status'] = -1
            result['message'] = e
            current_session['loginready'] = False
        else:
            result = inspection.result()
            current_session['loginready'] = inspection.ready
            current_session['name'] = result.get('membername', None)
        print('login result', result)
        return jsonify(result=result)

Marvin.register(index)



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-02-22 10:38:28
# @Last modified by:   Brian Cherinka
# @Last modified time: 2017-07-31 12:07:00

from __future__ import print_function, division, absolute_import
from marvin.web.controllers.galaxy import make_nsa_dict
from marvin.tools.cube import Cube
from marvin.tests.web.conftest import Page
from marvin.tests import marvin_test_if
import pytest


@pytest.fixture()
def cube(galaxy, mode):
    cube = Cube(plateifu=galaxy.plateifu, mode=mode)
    cube.exp_nsa_plotcols = galaxy.nsa_data
    return cube


@pytest.fixture()
def params(galaxy):
    return {'release': galaxy.release}


@pytest.mark.parametrize('page', [('galaxy_page', 'Galaxy:index')], ids=['galaxy'], indirect=True)
class TestGalaxyPage(object):

    def test_assert_galaxy_template_used(self, page, get_templates):
        page.load_page('get', page.url)
        assert '' == page.data
        template, context = get_templates[0]
        assert 'galaxy.html' == template.name, 'Template used should be galaxy.html'


@pytest.mark.parametrize('page', [('galaxy_page', 'initnsaplot')], ids=['initnsa'], indirect=True)
class TestNSA(object):

    #@marvin_test_if(mark='skip', cube=dict(nsa=[None]))
    def test_nsadict_correct(self, page, cube):
        nsa, cols = make_nsa_dict(cube.nsa)
        for value in cube.exp_nsa_plotcols.values():
            assert set(value.keys()).issubset(set(cols))
            page.assert_dict_contains_subset(value, nsa)
            page.assertListIn(value.keys(), cols)

    def test_initnsa_method_not_allowed(self, page, params, get_templates):
        page.load_page('get', page.url, params=params)
        template, context = get_templates[0]
        assert template.name == 'errors/method_not_allowed.html'

    def test_initnsa_no_plateifu(self, page, get_templates):
        errmsg = 'Field may not be null.'
        page.load_page('post', page.url)
        template, context = get_templates[0]
        page.route_no_valid_webparams(template, context, 'plateifu', reqtype='post', errmsg=errmsg)





# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-05-24 18:27:50
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-05-28 15:17:28

from __future__ import print_function, division, absolute_import
from marvin.tools.query.query_utils import query_params, QueryParameter
import pytest
import itertools


@pytest.fixture(scope='session')
def data():
    groups = ['Metadata', 'Spaxel Metadata', 'Emission Lines', 'Kinematics', 'Spectral Indices', 'NSA Catalog']
    spaxelparams = ['spaxelprop.x', 'spaxelprop.y', 'spaxelprop.spx_snr']
    specindparams = ['spaxelprop.specindex_d4000']
    nsaparams = ['nsa.z', 'nsa.elpetro_logmass', 'nsa.elpetro_th50_r']
    data = {'groups': groups, 'spaxelmeta': spaxelparams, 'nsa': nsaparams, 'specind': specindparams}

    return data


class TestGroupList(object):

    def test_list_groups(self, data):
        groups = query_params.list_groups()
        assert data['groups'] == groups, 'groups should be the same'

    @pytest.mark.parametrize('name, result',
                             [('metadata', 'Metadata'),
                              ('spaxelmeta', 'Spaxel Metadata'),
                              ('emission', 'Emission Lines'),
                              ('kin', 'Kinematics'),
                              ('nsacat', 'NSA Catalog')])
    def test_get_group(self, name, result):
        group = query_params[name]
        assert result == group.name

    @pytest.mark.parametrize('name',
                             [('nsa catalog'),
                              ('NSA catalog'),
                              ('catalognsa'),
                              ('nsa--!catalgue'),
                              ('nsacat')])
    def test_different_keys(self, name):
        group = query_params[name]
        assert group.name == 'NSA Catalog'

    @pytest.mark.parametrize('groups', [(None), (['nsa']), (['specind', 'nsa']),
                                        (['spaxelmeta', 'specind'])])
    def test_list_params(self, data, groups):
        params = query_params.list_params(groups=groups)
        if not groups:
            myparams = data['spaxelmeta'] + data['nsa'] + data['specind']
        else:
            paramlist = [data[g] for g in groups]
            myparams = list(itertools.chain.from_iterable(paramlist))
        assert set(myparams).issubset(set(params))

    def test_raises_keyerror(self):
        errmsg = "meta is too ambiguous."
        with pytest.raises(KeyError) as cm:
            group = query_params['meta']
        assert cm.type == KeyError
        assert errmsg in str(cm.value)

    @pytest.mark.parametrize('name, result',
                             [('coord', None),
                              pytest.mark.xfail(raises=KeyError)(('meta', None))])
    def test_failure(self, name, result):
        res = query_params[name]
        assert result == res, 'result should be the same as res'


class TestParamList(object):

    @pytest.mark.parametrize('group, name, count', [('specind', 'Spectral Indices', 1),
                                                    ('kin', 'Kinematics', 6)])
    def test_get_paramgroup(self, group, name, count):
        assert group in query_params
        mygroup = query_params[group]
        othergroup = group == query_params
        assert type(mygroup) == type(othergroup)
        assert mygroup.name == othergroup.name
        assert mygroup.name == name
        assert len(mygroup) == count

    @pytest.mark.parametrize('group, param, name', [('specind', 'd4000', 'specindex_d4000')])
    def test_get_param(self, group, param, name):
        assert group in query_params
        assert param in query_params[group]
        myparam = param == query_params[group]
        assert myparam.name == name

    @pytest.mark.parametrize('group, param, ltype, expname',
                             [('kin', 'havel', None, None),
                              ('kin', 'havel', 'full', 'spaxelprop.emline_gvel_ha_6564'),
                              ('kin', 'havel', 'short', 'havel'),
                              ('kin', 'havel', 'display', 'Halpha Velocity')])
    def test_list_params(self, group, param, ltype, expname):
        kwargs = {ltype: True} if ltype else {}
        params = query_params[group].list_params(**kwargs)
        if not ltype:
            assert isinstance(params[0], QueryParameter)
        else:
            assert params[0] == expname

    @pytest.mark.parametrize('group, params, expset',
                             [('metadata', ['ra', 'dec'], ['cube.ra', 'cube.dec'])])
    def test_list_subset(self, group, params, expset):
        subset = query_params[group].list_params(params, full=True)
        assert set(expset) == set(subset)

    @pytest.mark.parametrize('groups, params1, params2, expset',
                             [(('metadata', 'nsa'), ['ra', 'dec'], ['z', 'absmag_g_r'],
                               ['cube.ra', 'cube.dec', 'nsa.z', 'nsa.elpetro_absmag_g_r'])])
    def test_join_two_list(self, groups, params1, params2, expset):
        group1, group2 = groups
        g1 = query_params[group1]
        g2 = query_params[group2]
        mylist = g1.list_params(params1, full=True) + g2.list_params(params2, full=True)
        assert set(expset) == set(mylist)

    def test_raises_keyerror(self):
        errmsg = "emline_gflux is too ambiguous."
        with pytest.raises(KeyError) as cm:
            param = query_params['emission']['emline_gflux']
        assert cm.type == KeyError
        assert errmsg in str(cm.value)


class TestQueryParams(object):

    @pytest.mark.parametrize('group, param, full, name, short, display',
                             [('nsa', 'z', 'nsa.z', 'z', 'z', 'Redshift'),
                              ('metadata', 'ra', 'cube.ra', 'ra', 'ra', 'RA'),
                              ('emission', 'gflux_ha', 'spaxelprop.emline_gflux_ha_6564', 'emline_gflux_ha_6564', 'haflux', 'Halpha Flux'),
                              ('spaxelmeta', 'x', 'spaxelprop.x', 'x', 'x', 'Spaxel X')])
    def test_query_param(self, group, param, full, name, short, display):
        par = query_params[group][param]
        assert par.full == full
        assert par.name == name
        assert par.short == short
        assert par.display == display




#!/usr/bin/env python
# encoding: utf-8
"""

exceptions.py

Licensed under a 3-clause BSD license.

Revision history:
    13 Feb 2016 J. Sánchez-Gallego
      Initial version

"""

from __future__ import division
from __future__ import print_function


__all__ = ['BrainError', 'BrainUserWarning', 'BrainSkippedTestWargning',
           'BrainNotImplemented', 'BrainApiAuthError']


class BrainError(Exception):
    pass


class BrainApiAuthError(BrainError):
    pass


class BrainNotImplemented(BrainError):
    """A Brain exception for not yet implemented features."""

    def __init__(self, message=None):

        message = 'This feature is not implemented yet.' \
            if not message else message

        super(BrainNotImplemented, self).__init__(message)


class BrainMissingDependence(BrainError):
    """A custom exception for missing dependences."""
    pass


class BrainWarning(Warning):
    """Base warning for Brain."""
    pass


class BrainUserWarning(UserWarning, BrainWarning):
    """The primary warning class."""
    pass


class BrainSkippedTestWargning(BrainUserWarning):
    """A warning for when a test is skipped."""
    pass



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-05-07 15:58:27
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-06-26 14:20:20

from __future__ import print_function, division, absolute_import
from marvin.api.base import BaseView
import pytest


@pytest.fixture(autouse=True)
def baseview():
    baseview = BaseView()
    yield baseview


class TestBase(object):

    def test_reset_results(self, baseview):
        baseview.results = {'key1': 'value1'}
        baseview.reset_results()
        desired = {'data': None, 'status': -1, 'error': None, 'traceback': None}
        assert baseview.results == desired, 'baseview results should be the same as desired'

    def test_update_results(self, baseview):
        new_results = {'key1': 'value1'}
        baseview.update_results(new_results)
        desired = {'data': None, 'status': -1, 'error': None, 'key1': 'value1', 'traceback': None}
        assert baseview.results == desired, 'baseview results should be the same as desired'

    def test_reset_status(self, baseview):
        baseview.results['status'] = 42
        baseview.reset_status()
        assert baseview.results['status'] == -1

    def test_add_config(self, baseview, release, mode):
        baseview.add_config()
        desired = {'data': None, 'status': -1, 'error': None, 'traceback': None,
                   'utahconfig': {'release': 'MPL-5', 'mode': 'local'}}
        assert baseview.results == desired

    def test_after_request_return_response(self, baseview):
        name = 'test_name'
        req = 'test_request'
        actual = baseview.after_request(name, req)
        desired = 'test_request'
        assert actual == desired

    def test_after_request_reset_results(self, baseview):
        name = 'test_name'
        req = 'test_request'
        baseview.after_request(name, req)
        desired = {'data': None, 'status': -1, 'error': None, 'traceback': None}
        assert baseview.results == desired




#!/usr/bin/env python
# encoding: utf-8
#
# test_bpt_pytest.py
#
# Created by José Sánchez-Gallego on 10 Feb 2017.
# Converted to pytest by Brett Andrews on 15 Jun 2017.


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from matplotlib import pyplot as plt
import numpy as np
import pytest

from marvin.tools.maps import Maps
from marvin.tests import marvin_test_if_class
from marvin.core.exceptions import MarvinDeprecationWarning


@pytest.fixture()
def maps(galaxy, mode):
    maps = Maps(plateifu=galaxy.plateifu, mode=mode)
    maps.bptsums = galaxy.bptsums if hasattr(galaxy, 'bptsums') else None
    yield maps
    maps = None


@marvin_test_if_class(mark='skip', maps=dict(release=['MPL-4']))
class TestBPT(object):

    mechanisms = ['sf', 'comp', 'agn', 'seyfert', 'liner', 'invalid', 'ambiguous']

    @pytest.mark.parametrize('useoi', [(True), (False)], ids=['withoi', 'nooi'])
    def test_bpt(self, maps, useoi):
        if maps.bptsums is None:
            pytest.skip('no bpt data found in galaxy test data')

        bptflag = 'nooi' if useoi is False else 'global'

        masks, figure = maps.get_bpt(show_plot=False, return_figure=True, use_oi=useoi)
        assert isinstance(figure, plt.Figure)

        for mech in self.mechanisms:
            assert mech in masks.keys()
            assert np.sum(masks[mech]['global']) == maps.bptsums[bptflag][mech]

    def test_bpt_diffsn(self, maps):
        if maps.bptsums is None:
            pytest.skip('no bpt data found in galaxy test data')

        masks, figure = maps.get_bpt(show_plot=False, return_figure=True, use_oi=True, snr_min=5)
        assert isinstance(figure, plt.Figure)

        for mech in self.mechanisms:
            assert mech in masks.keys()
            assert np.sum(masks[mech]['global']) == maps.bptsums['snrmin5'][mech]

    def test_bpt_oldsn(self, maps):
        if maps.bptsums is None:
            pytest.skip('no bpt data found in galaxy test data')

        with pytest.warns(MarvinDeprecationWarning) as record:
            masks = maps.get_bpt(snr=5, return_figure=False, show_plot=False)
        assert len(record) == 1
        assert record[0].message.args[0] == "snr is deprecated. Use snr_min instead. snr will be removed in a future version of marvin"

        for mech in self.mechanisms:
            assert mech in masks.keys()
            assert np.sum(masks[mech]['global']) == maps.bptsums['snrmin5'][mech]




from collections import Iterable, Mapping

import six
from sqlalchemy import Column
from sqlalchemy.orm.attributes import InstrumentedAttribute
from wtforms import ValidationError


class Unique(object):
    """Checks field values unicity against specified table fields.

    :param column:
        InstrumentedAttribute object, eg. User.name, or
        Column object, eg. user.c.name, or
        a field name, eg. 'name' or
        a tuple of InstrumentedAttributes, eg. (User.name, User.email) or
        a dictionary mapping field names to InstrumentedAttributes, eg.
        {
            'name': User.name,
            'email': User.email
        }
    :param get_session:
        A function that returns a SQAlchemy Session. This parameter is not
        needed if the given model supports Flask-SQLAlchemy styled query
        parameter.
    :param message:
        The error message.
    """
    field_flags = ('unique', )

    def __init__(self, column, get_session=None, message=None):
        self.column = column
        self.message = message
        self.get_session = get_session

    @property
    def query(self):
        self._check_for_session(self.model)
        if self.get_session:
            return self.get_session().query(self.model)
        elif hasattr(self.model, 'query'):
            return getattr(self.model, 'query')
        else:
            raise Exception(
                'Validator requires either get_session or Flask-SQLAlchemy'
                ' styled query parameter'
            )

    def _check_for_session(self, model):
        if not hasattr(model, 'query') and not self.get_session:
            raise Exception('Could not obtain SQLAlchemy session.')

    def _syntaxes_as_tuples(self, form, field, column):
        """Converts a set of different syntaxes into a tuple of tuples"""
        if isinstance(column, six.string_types):
            return ((column, getattr(form.Meta.model, column)),)
        elif isinstance(column, Mapping):
            return tuple(
                (x[0], self._syntaxes_as_tuples(form, field, x[1])[0][1])
                for x in column.items()
            )
        elif isinstance(column, Iterable):
            return tuple(
                self._syntaxes_as_tuples(form, field, x)[0]
                for x in column
            )
        elif isinstance(column, (Column, InstrumentedAttribute)):
            return ((column.key, column),)
        else:
            raise TypeError("Invalid syntax for column")

    def __call__(self, form, field):
        columns = self._syntaxes_as_tuples(form, field, self.column)
        self.model = columns[0][1].class_
        query = self.query
        for field_name, column in columns:
            query = query.filter(column == form[field_name].data)
        obj = query.first()

        if not hasattr(form, '_obj'):
            raise Exception(
                "Couldn't access Form._obj attribute. Either make your form "
                "inherit WTForms-Alchemy ModelForm or WTForms-Components "
                "ModelForm or make this attribute available in your form."
            )

        if obj and not form._obj == obj:
            if self.message is None:
                self.message = field.gettext(u'Already exists.')
            raise ValidationError(self.message)



from distutils.version import LooseVersion

import wtforms
from pytest import raises
from wtforms import Form

from tests import ModelFormTestCase
from wtforms_alchemy import (
    FormGenerator,
    model_form_factory,
    UnknownConfigurationOption
)


class TestModelFormFactory(ModelFormTestCase):
    def test_supports_parameter_overriding(self):
        self.init()

        class MyFormGenerator(FormGenerator):
            pass

        defaults = {
            'all_fields_optional': True,
            'only_indexed_fields': True,
            'include_primary_keys': True,
            'include_foreign_keys': True,
            'strip_string_fields': True,
            'include_datetimes_with_default': True,
            'form_generator': True,
            'date_format': '%d-%m-%Y',
            'datetime_format': '%Y-%m-%dT%H:%M:%S',
        }
        ModelForm = model_form_factory(Form, **defaults)
        for key, value in defaults.items():
            assert getattr(ModelForm.Meta, key) == value

    def test_throws_exception_for_unknown_configuration_option(self):
        self.init()

        class SomeForm(Form):
            pass

        defaults = {
            'unknown': 'something'
        }
        with raises(UnknownConfigurationOption):
            model_form_factory(SomeForm, **defaults)

    def test_supports_custom_base_class_with_model_form_factory(self):
        self.init()

        class SomeForm(Form):
            pass

        class TestCustomBase(model_form_factory(SomeForm)):
            class Meta:
                model = self.ModelTest

        assert isinstance(TestCustomBase(), SomeForm)

    def test_url_validator(self):
        form = model_form_factory(url_validator=None)
        assert form.Meta.url_validator is None

    def test_email_validator(self):
        form = model_form_factory(email_validator=None)
        assert form.Meta.email_validator is None

    def test_length_validator(self):
        form = model_form_factory(length_validator=None)
        assert form.Meta.length_validator is None

    def test_number_range_validator(self):
        form = model_form_factory(number_range_validator=None)
        assert form.Meta.number_range_validator is None

    def test_date_range_validator(self):
        form = model_form_factory(date_range_validator=None)
        assert form.Meta.date_range_validator is None

    def test_time_range_validator(self):
        form = model_form_factory(time_range_validator=None)
        assert form.Meta.time_range_validator is None

    def test_optional_validator(self):
        form = model_form_factory(optional_validator=None)
        assert form.Meta.optional_validator is None

    def test_unique_validator(self):
        form = model_form_factory(unique_validator=None)
        assert form.Meta.unique_validator is None

    def test_class_meta_wtforms2(self):
        if LooseVersion(wtforms.__version__) < LooseVersion('2'):
            return  # skip test for wtforms < 2

        self.init()

        class SomeForm(Form):
            class Meta:
                locales = ['fr']
                foo = 9

        class OtherForm(SomeForm):
            class Meta:
                pass

        class TestCustomBase(model_form_factory(SomeForm)):
            class Meta:
                model = self.ModelTest

        form = TestCustomBase()
        other_form = OtherForm()
        assert isinstance(form.meta, wtforms.meta.DefaultMeta)
        assert form.meta.locales == ['fr']
        assert hasattr(form.meta, 'model')
        assert hasattr(form.meta, 'csrf')

        assert form.meta.foo == 9
        # Create a side effect on the base meta.
        SomeForm.Meta.foo = 12
        assert other_form.meta.foo == 12
        assert form.meta.foo == 12



#!/usr/bin/env python
# encoding: utf-8

'''
Created by Brian Cherinka on 2016-04-26 09:20:35
Licensed under a 3-clause BSD license.

Revision History:
    Initial Version: 2016-04-26 09:20:35 by Brian Cherinka
    Last Modified On: 2016-04-26 09:20:35 by Brian
'''

from sqlalchemy import create_engine, MetaData
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.event import listen
from sqlalchemy.pool import Pool
from marvin.core import caching_query
from hashlib import md5
from dogpile.cache.region import make_region
import os

# DOGPILE CACHING SETUP

# dogpile cache regions.  A home base for cache configurations.
regions = {}

dogpath = os.environ.get('MANGA_SCRATCH_DIR', None)
if dogpath and os.path.isdir(dogpath):
    dogpath = dogpath
else:
    dogpath = os.path.expanduser('~')

dogroot = os.path.join(dogpath, 'dogpile_data')
if not os.path.isdir(dogroot):
    os.makedirs(dogroot)


# make an nsa region
def make_nsa_region(name):
    reg = make_region(key_mangler=md5_key_mangler).configure(
            'dogpile.cache.dbm',
            expiration_time=3600,
            arguments={'filename': os.path.join(dogroot, '{0}_cache.dbm'.format(name))}
        )
    return reg


# db hash key
def md5_key_mangler(key):
    """Receive cache keys as long concatenated strings;
    distill them into an md5 hash.

    """
    return md5(key.encode('ascii')).hexdigest()

# configure the "default" cache region.
regions['default'] = make_region(
            # the "dbm" backend needs string-encoded keys
            key_mangler=md5_key_mangler
        ).configure(
        # using type 'file' to illustrate
        # serialized persistence.  Normally
        # memcached or similar is a better choice
        # for caching.
        # 'dogpile.cache.dbm',  # file-based backend
        'dogpile.cache.memcached',  # memcached-based backend
        expiration_time=3600,
        arguments={
            'url': "127.0.0.1:11211"  # memcached option
            # "filename": os.path.join(dogroot, "cache.dbm") # file option
        }
    )
regions['nsa_mpl5'] = make_nsa_region('nsa_mpl5')
regions['nsa_mpl4'] = make_nsa_region('nsa_mpl4')


def clearSearchPathCallback(dbapi_con, connection_record):
    '''
    When creating relationships across schema, SQLAlchemy
    has problems when you explicitly declare the schema in
    ModelClasses and it is found in search_path.

    The solution is to set the search_path to "$user" for
    the life of any connection to the database. Since there
    is no (or shouldn't be!) schema with the same name
    as the user, this effectively makes it blank.

    This callback function is called for every database connection.

    For the full details of this issue, see:
    http://groups.google.com/group/sqlalchemy/browse_thread/thread/88b5cc5c12246220

    dbapi_con - type: psycopg2._psycopg.connection
    connection_record - type: sqlalchemy.pool._ConnectionRecord
    '''
    cursor = dbapi_con.cursor()
    cursor.execute('SET search_path TO "$user",functions')
    dbapi_con.commit()

listen(Pool, 'connect', clearSearchPathCallback)


class DatabaseConnection(object):
    '''This class defines an object that makes a connection to a database.
       The "DatabaseConnection" object takes as its parameter the SQLAlchemy
       database connection string.

       This class is best called from another class that contains the
       actual connection information (so that it can be reused for different
       connections).

       This class implements the singleton design pattern. The first time the
       object is created, it *requires* a valid database connection string.
       Every time it is called via:

       db = DatabaseConnection()

       the same object is returned and contains the connection information.
    '''
    _singletons = dict()

    def __new__(cls, database_connection_string=None, expire_on_commit=True):
        """This overrides the object's usual creation mechanism."""

        if cls not in cls._singletons:
            assert database_connection_string is not None, "A database connection string must be specified!"
            cls._singletons[cls] = object.__new__(cls)

            # ------------------------------------------------
            # This is the custom initialization
            # ------------------------------------------------
            me = cls._singletons[cls]  # just for convenience (think "self")

            me.database_connection_string = database_connection_string

            # change 'echo' to print each SQL query (for debugging/optimizing/the curious)
            me.engine = create_engine(me.database_connection_string, echo=False, pool_size=10, pool_recycle=1800)

            me.metadata = MetaData()
            me.metadata.bind = me.engine
            me.Base = declarative_base(bind=me.engine)
            me.Session = scoped_session(sessionmaker(bind=me.engine, autocommit=True,
                                                     query_cls=caching_query.query_callable(regions),
                                                     expire_on_commit=expire_on_commit))
            # ------------------------------------------------

        return cls._singletons[cls]





#!/usr/bin/env python
# encoding: utf-8

'''
Licensed under a 3-clause BSD license.

Revision History:
    Initial Version: 2016-02-23 14:13:28 by Brian Cherinka
    2016-02-23 - made some test sample forms with wtforms+sqlalchemy - B. Cherinka
    2016-03-02 - generalized the form to build all forms - B. Cherinka
'''

from __future__ import print_function
from __future__ import division
import sys
from marvin import marvindb, config
from marvin.core.exceptions import MarvinError, MarvinUserWarning
from collections import defaultdict
from wtforms import StringField, Field, FieldList, validators
from wtforms import SelectMultipleField, ValidationError, SubmitField
from wtforms.widgets import TextInput
from wtforms_alchemy import model_form_factory
from sqlalchemy.inspection import inspect as sa_inspect
from sqlalchemy.ext.hybrid import hybrid_property, hybrid_method
from sqlalchemy.orm.attributes import InstrumentedAttribute
import re
import warnings

# flask_wtf does not work locally - OUTSIDE APPLICATON CONTEXT; need some kind of toggle for web version and not???
if config._inapp:
    from flask_wtf import Form
else:
    from wtforms import Form

__all__ = ['MarvinForm']


def tree():
    return defaultdict(tree)


# Base form class
BaseModelForm = model_form_factory(Form)


class ModelForm(BaseModelForm):
    ''' sub class a new ModelForm so it works with Flask-WTF in APP mode;  for auto CSRF tokens...who knows...'''
    @classmethod
    def get_session(self):
        return marvindb.session

''' Builds a dictionary for modelclasses with key ClassName and value SQLalchemy model class ; '''
modelclasses = marvindb.buildUberClassDict()


# class factory
def formClassFactory(name, model, baseclass):
    ''' Generates a new WTForm class based on SQLalchemy model class.  Each class contains as attributes
        Meta = a class called Meta.  Meta.model contains the SQLalchemy ModelClass
        data = a dictionary of parameters: form input that gets mapped to the sqlalchemy parameter
        errors = a dictionary of errors returned by invalid form validation
        validate = a method to validate all elements in this form
        parameter_X = a WTForm Field mapped to respective sqlalchemy table column

        e.g.
        The ModelClass IFUDesign mapped to mangadatadb.ifu_design sql table gets transformed into
        WTForm IFUDesignForm, with IFUDesignForm.Meta.model = marvin.db.models.DataModelClasses.IFUDesign
    '''

    Meta = type('Meta', (object,), {'model': model})
    newclass = type(name, (baseclass,), {'Meta': Meta})
    return newclass

# build a wtform select field for operators; tested but no longer used ; can't seem to attach operator field to every individual parameter
opdict = {'le': '<=', 'ge': '>=', 'gt': '>', 'lt': '<', 'ne': '!=', 'eq': '='}
ops = [(key, val) for key, val in opdict.items()]
# operator = SelectField(u'Operator', choices=ops)


class ParamFxnLookupDict(dict):
    ''' Parameter function lookup for new function expressions
    '''

    def __getitem__(self, key):

        lowkey = key.lower()
        mykeys = list(self.keys())

        inkey = lowkey in mykeys
        if not inkey:
            raise KeyError('{0} does not match any column.'.format(lowkey))
        else:
            keycount = mykeys.count(lowkey)
            if keycount > 1:
                raise KeyError('{0} matches multiple parameters in the lookup'
                               ' table'.format(lowkey))
            else:
                return dict.__getitem__(self, lowkey)


class ParamFormLookupDict(dict):

    def __init__(self, **kwargs):
        self.allspaxels = kwargs.get('allspaxels', None)
        self._release = kwargs.get('release', config.release)
        self._init_table_shortcuts()
        self._init_name_shortcuts()

    def __getitem__(self, key):
        """Checks if `key` is a unique column name and return the value."""

        # Init the shortcuts
        self._init_table_shortcuts()
        self._init_name_shortcuts()

        # Applies shortcuts
        keySplits = self._apply_shortcuts(key)

        # Get key matches
        matches = self._get_matches(keySplits)

        # Check DAP Junk keys
        matches = self._check_for_junk(matches)

        # Return matched key
        if len(matches) == 0:
            # No matches. This returns the standards KeyError from dict
            raise KeyError('{0} does not match any column.'.format(key))
        elif len(matches) == 1:
            # One match: returns the form.
            return dict.__getitem__(self, matches[0])
        else:
            # Multiple results. Raises a custom error.
            raise KeyError(
                '{0} matches multiple parameters in the lookup table: {1}'
                .format(key, ', '.join(matches)))

    def _check_for_junk(self, matches):
        ''' check for Junk matches and return the correct key '''
        isdapjunk = any(['mangadapdb.spaxelprop' in m for m in matches])
        if isdapjunk:
            junkmatches = [m for m in matches if (self.allspaxels and 'clean'
                           not in m) or (not self.allspaxels and 'clean' in m)]
            keySplits = self._apply_shortcuts(junkmatches[0])
            matches = self._get_matches(keySplits)
        return matches

    def mapToColumn(self, keys):
        """Returns the model class column in the WFTForm."""

        if not isinstance(keys, (list, tuple)):
            str_types = [str]
            if sys.version_info[0] < 3:
                str_types.append(unicode)
            if isinstance(keys, tuple(str_types)):
                keys = [keys]
            else:
                keys = list(keys)

        columns = []
        for key in keys:
            keySplits = self._apply_shortcuts(key)
            matches = self._get_matches(keySplits)
            matches = self._check_for_junk(matches)
            if len(matches) == 0:
                raise KeyError('{0} does not match any column.'.format(key))
            elif len(matches) == 1:
                key = matches[0]
            else:
                raise KeyError('{0} matches multiple parameters \
                    in the lookup table: {1}'.format(key, ', '.join(matches)))
            wtfForm = self[key]
            column = key.split('.')[-1]
            columns.append(getattr(wtfForm.Meta.model, column))

        if len(columns) == 1:
            return columns[0]
        else:
            return columns

    def _init_table_shortcuts(self):
        ''' initialize the table shortcuts '''

        self._tableShortcuts = {'ifu': 'ifudesign', 'cube_header_keyword': 'fits_header_keyword',
                                'cube_header_value': 'fits_header_value', 'maps_header_keyword': 'header_keyword',
                                'maps_header_value': 'header_value'}
        self._set_junk_shortcuts()

    def _init_name_shortcuts(self):
        ''' initialize the name shortcuts '''
        self._nameShortcuts = {'haflux': 'emline_gflux_ha_6564',
                               'g_r': 'elpetro_mag_g_r',
                               'abs_g_r': 'elpetro_absmag_g_r'}

    def _apply_shortcuts(self, key):
        ''' Apply the shortcuts to the key '''

        # Gets the paths that match the key
        keySplits = key.split('.')

        # Applies table shortcuts
        if len(keySplits) >= 2 and keySplits[-2] in self._tableShortcuts:
            keySplits[-2] = self._tableShortcuts[keySplits[-2]]

        # Applies name shortcuts
        if len(keySplits) >= 1 and keySplits[-1] in self._nameShortcuts:
            keySplits[-1] = self._nameShortcuts[keySplits[-1]]

        return keySplits

    def _get_real_key(self, key):
        ''' Returns the real full key given some shortcuts '''
        keySplits = self._apply_shortcuts(key)
        return '.'.join(keySplits)

    def _set_junk_shortcuts(self):
        ''' Sets the DAP spaxelprop shortcuts based on MPL '''

        newmpls = [m for m in config._mpldict.keys() if m >= 'MPL-4']
        spaxname = 'spaxelprop' if self.allspaxels else 'cleanspaxelprop'
        if '4' in self._release:
            dapcut = {'spaxelprop{0}'.format(m.split('-')[1]): spaxname for m in newmpls}
            dapcut.update({'spaxelprop': spaxname})
        else:
            mdigit = self._release.split('-')[1]
            dapcut = {'spaxelprop{0}'.format(m.split('-')[1]): '{0}{1}'.format(spaxname, mdigit) for m in newmpls}
            dapcut.update({'spaxelprop': '{0}{1}'.format(spaxname, mdigit)})

        # add junk shortcuts
        junkcuts = {k.replace('spaxelprop', 'junk'): v for k, v in dapcut.items()}
        dapcut.update(junkcuts)

        # update the main dictionary
        self._tableShortcuts.update(dapcut)

    def _get_matches(self, keySplits):
        ''' Get the matches from a set of key splits '''
        matches = [path for path in self
                   if all([keySplits[-1 - ii] == path.split('.')[-1 - ii]
                           for ii in range(len(keySplits))])]
        return matches


# Custom [Better]TagList Field for Search Parameter Box
class TagListField(Field):
    widget = TextInput()

    def _value(self):
        if self.data:
            return u', '.join(self.data)
        else:
            return u''

    def process_formdata(self, valuelist):
        if valuelist:
            self.data = [x.strip() for x in valuelist[0].split(',') if x]
        else:
            self.data = []


class BetterTagListField(TagListField):
    def __init__(self, label='', validators=None, remove_duplicates=True, **kwargs):
        super(BetterTagListField, self).__init__(label, validators, **kwargs)
        self.remove_duplicates = remove_duplicates

    def process_formdata(self, valuelist):
        super(BetterTagListField, self).process_formdata(valuelist)
        if self.remove_duplicates:
            self.data = list(self._remove_duplicates(self.data))

    @classmethod
    def _remove_duplicates(cls, seq):
        """Remove duplicates in a case insensitive, but case preserving manner"""
        dups = {}
        for item in seq:
            if item.lower() not in dups:
                dups[item.lower()] = True
                yield item


# Custom validator for MainForm
class ValidOperand(object):
    def __init__(self, opstring='[<>=]', message=None):
        self.opstring = opstring
        if not message:
            message = u'Field must contain at least a valid operand of {0}.'.format(self.opstring)
        self.message = message

    def __call__(self, form, field):
        infield = re.search(self.opstring, field.data)
        if not infield:
            raise ValidationError(self.message)


class SearchForm(Form):
    ''' Main Search Level WTForm for Marvin '''

    searchbox = StringField("<a target='_blank' href='https://sdss-marvin.readthedocs.io/en/stable/tools/query/query_using.html'>Input Search Filter</a>",
        [validators.Length(min=3, message='Input must have at least 3 characters'),
                            validators.DataRequired(message='Input filter string required'),
                            ValidOperand('[<>=between]', message='Input must contain a valid operand.')])
    returnparams = SelectMultipleField("<a target='_blank' href='https://api.sdss.org/doc/manga/marvin/query_params.html'>Return Parameters</a>")
    submitsearch = SubmitField('Search')


class MarvinForm(object):
    ''' Core Marvin Form object. '''

    def __init__(self, *args, **kwargs):
        ''' Initialize the obect.  On init, generates all the WTForms from the ModelClasses.

        _param_form_lookup = dictionary of all modelclass parameters of form {'SQLalchemy ModelClass parameter name': WTForm Class}
        '''

        self._release = kwargs.get('release', config.release)
        self._modelclasses = marvindb.buildUberClassDict(release=self._release)
        self._param_form_lookup = ParamFormLookupDict(**kwargs)
        self._param_fxn_lookup = ParamFxnLookupDict()
        self._paramtree = tree()
        self._generateFormClasses(self._modelclasses)
        self._generateFxns()
        self.SearchForm = SearchForm
        self._cleanParams(**kwargs)

    def _generateFormClasses(self, classes):
        ''' Loops over all ModelClasses and generates a new WTForm class.  New form classes are named as [ModelClassName]Form.
            Sets the new form as an attribute on MarvinForm.  Also populates the _param_to_form_lookup dictonary with
            all ModelClass/WTForm parameters and their corresponding forms.

            e.g.  _param_form_lookup['name'] = marvin.tools.query.forms.IFUDesignForm
        '''

        for key, val in classes.items():
            classname = '{0}Form'.format(key)
            try:
                newclass = formClassFactory(classname, val, ModelForm)
            except Exception as e:
                warnings.warn('class {0} not Formable'.format(key), MarvinUserWarning)
            else:
                self.__setattr__(classname, newclass)
                self._loadParams(newclass)

    def _loadParams(self, newclass):
        ''' Loads all parameters from wtforms into a dictionary with
            key, value = {'parameter_name': 'parent WTForm name'}.
            Ignores hidden attributes and the Meta class
        '''

        model = newclass.Meta.model
        schema = model.__table__.schema
        tablename = model.__table__.name

        mapper = sa_inspect(model)
        for key, item in mapper.all_orm_descriptors.items():
            if isinstance(item, hybrid_property) or \
               isinstance(item, hybrid_method):
                key = key
            elif isinstance(item, InstrumentedAttribute):
                key = item.key
            else:
                continue

            lookupKeyName = schema + '.' + tablename + '.' + key
            self._param_form_lookup[lookupKeyName] = newclass
            self._paramtree[newclass.Meta.model.__name__][key]

    def callInstance(self, form, params=None, **kwargs):
        ''' Creates an instance of a specified WTForm.  '''
        return form(**params) if params else form(**kwargs)

    def _generateFxns(self):
        ''' Generate the fxn dictionary

            TODO: make this general and not hard-coded

            The key is the function name used in the query syntax.
            The value is the method call that lives in Query

        '''
        self._param_fxn_lookup['npergood'] = 'getPercent'

    def _getDapKeys(self):
        ''' Returns the DAP keys from the Junk tables
        '''
        dapkeys = [k for k in self._param_form_lookup.keys() if 'mangadapdb.spaxelprop' in k]
        dapkeys.sort()
        return dapkeys

    def _cleanParams(self, **kwargs):
        ''' Clean up the parameter-form lookup dictionary '''

        # remove keys for pk, mangadatadb.sample, test_, and cube_header
        new = ParamFormLookupDict(**kwargs)
        for k, v in self._param_form_lookup.items():
            if 'pk' not in k and \
               'mangadatadb.sample' not in k and \
               'test_' not in k and \
               'cube_header' not in k:
                new[k] = v

        # make new dictionary
        self._param_form_lookup = new



import numpy as np
import os
import decimal
import datetime

# General utilities

__all__ = []


def getDbMachine():
    ''' Get the machine that the app is running on.

    This determines correct database and app configuration '''
    # Get machine
    machine = os.environ.get('HOSTNAME', None)

    # Check if localhost or not
    try:
        localhost = bool(os.environ['MANGA_LOCALHOST'])
    except:
        localhost = machine == 'manga'

    # Check if Utah or not
    try:
        utah = os.environ['UUFSCELL'] == 'kingspeak.peaks'
    except:
        utah = None
    # Check if sas-vm or not
    sasvm = 'sas-vm' in machine if machine else None

    # Set the dbconfig variable
    if localhost:
        return 'local'
    elif utah or sasvm:
        return 'utah'
    else:
        return None


def convertIvarToErr(ivar):
    ''' Converts a list of inverse variance into an a list of standard errors '''

    assert type(ivar) == list or type(ivar) == np.ndarray, \
        'Input ivar is not of type list or an Numpy ndarray'

    if type(ivar) == list:
        ivar = np.array(ivar)

    error = np.zeros(ivar.shape)
    notnull = ivar != 0.0
    error[notnull] = 1 / np.sqrt(ivar[notnull])
    error = list(error)
    return error


def alchemyencoder(obj):
    """JSON encoder function for SQLAlchemy special classes."""
    if isinstance(obj, datetime.date):
        return obj.isoformat()
    elif isinstance(obj, decimal.Decimal):
        return float(obj)



#!/usr/bin/env python
# encoding: utf-8
"""

modelPaths.py

Created by José Sánchez-Gallego on 18 Mar 2016.
Licensed under a 3-clause BSD license.

Revision history:
    18 Mar 2016 J. Sánchez-Gallego
      Initial version
    29 Mar 2016 - Removed default datadb dependency - B. Cherinka

"""

from __future__ import division
from __future__ import print_function
from sqlalchemy.ext.declarative import DeclarativeMeta
import itertools
import numpy as np
import networkx as nx


def isModel(model):
    """Return True if the input is a Model Class."""

    if (isinstance(model, DeclarativeMeta) and hasattr(model, '__table__') and
            model.__table__.name):
        return True
    else:
        False


def getModels(module):
    """Returns a list with all the ModelClasses for a module.

    E.g., for ``datadb`` it returns PipelineCompletionStatus, PipelineInfo,
    IFUDesign, Cube, Sample, etc.

    """

    models = []
    attrs = list(vars(module).keys())

    # Loops over all the attributes in the module and selects those that are
    # model classes.
    for attr in attrs:
        testModel = getattr(module, attr)
        if isModel(testModel):
            models.append(getattr(module, attr))

    return models


class ModelGraph(object):
    """Creates a `networkx` graph between the model classes.

    This class creates a graph in which each table in the input
    ``modelSchemas`` is a node, and the foreign keys relating them act as
    edges (links). The class includes some convenience methods to find the
    shortest path between two or more tables/nodes.

    The main purpose of the class is to help finding the necessary tables
    to join for a SQLalchemy query to work. For instance, one may want
    to perform a query such as ``'ifu.nfiber == 127 and nsa.nsa_z > 0.1'``.
    From that query only two model classes are retrieved:
    ``mangadatadb.IFUDesign`` and ``mangasampledb.Nsa``, which don't have a
    direct relationship. However, it is possible to perform this query if
    we add the following models to the join: ``mangadatadb.Cube``,
    ``mangasampledb.Target``. Creating a graph with all the nodes and
    relationships in the database makes finding the shortest path between
    two tables trivial.

    Parameters:
        modelSchemas (module or list of modules):
            A module or list of modules containing model classes.
            Each model class will be considered a node in the graph, and their
            relationships will define the edges between nodes.

    Example:
      >>> graph = ModelGraph([datadb, sampledb])
      >>> graph.join_models([IFUDesign, Nsa])
          [IFUDesign, Cube, Target, Nsa]

    """

    def __init__(self, modelSchemas):

        self.schemas = np.atleast_1d(modelSchemas)
        self.models = {}

        # Initialites a graph in which the tables (schema.tablename) will be
        # the nodes.
        self.graph = nx.Graph()

        # Creates the nodes
        for schema in self.schemas:
            self._createNodes(schema)

        # Creates the edges from the relationships between models.
        for schema in self.schemas:
            self._createEdges(schema)

    @property
    def nodes(self):
        """Shortcut to self.graph.nodes()."""
        return self.graph.nodes()

    @property
    def edges(self):
        """Shortcut to self.graph.edges()."""
        return self.graph.edges()

    @staticmethod
    def getTablePath(model):
        """From a model, returns the table path in the form schema.table."""

        tableName = model.__table__.name
        schemaName = model.__table__.schema
        fullPath = schemaName + '.' + tableName

        return fullPath

    def _createNodes(self, schema):
        """Creates the nodes for the tables in a schema."""

        # Gets the models in a schema module
        self.models[schema.__name__] = getModels(schema)

        # Adds the table name of each model as a node.
        for model in self.models[schema.__name__]:
            self.graph.add_node(self.getTablePath(model), model=model)

    def _createEdges(self, schema):
        """Creates the edges between nodes in the table graph."""

        # We loop over each model, get the foreign keys, and create an edge
        # between parent and child. Networkx will automatically ignore inverse
        # edges (i.e., if we add an edge between Cube and Wavelength, and later
        # add the inverse edge between Wavelength and Cube, the latter edge
        # will be ignored, as a previous link between the two nodes already
        # exists).
        for model in self.models[schema.__name__]:

            foreignKeys = list(model.__table__.foreign_keys)
            for fKey in foreignKeys:
                childSchemaName = fKey.column.table.schema
                childTableName = fKey.column.table.name
                childFullPath = childSchemaName + '.' + childTableName
                if childFullPath not in self.graph.nodes():
                    continue
                else:
                    parent = self.getTablePath(model)
                    self.graph.add_edge(parent, childFullPath)

    def getJoins(self, models, format_out='tables', nexus=None):
        """Returns a list all model classes needed to perform a join.

        Given a list of ``models``, finds the shortest join paths between any
        two elements in the list. Returns a list of all the tables or models
        that need to be joined to perform a query on ``models``.

        Parameters
        ----------
        models : list of model classes or tablenames
            A list of the tables or model classes to join. If tablenames,
            the full path, ``schema.table`` must be provided. The format of the
            input is determined automatically from the type of the first
            element.
        format_out : string
            Defines the type of elements in the returned list. If ``'models'``,
            the returned elements will be model classes, if ``'tables'`` they
            will be the corresponding table paths (schema.table).
        nexus : string, model class, or None
            If None, the method will find the paths between each combination of
            two elements in the input ``model`` list. If a table is provided
            (either as a table path or as a model class), the returned list
            will be the shorted path between `nexus` and each on of the tables
            in ``models``. The ``nexus`` table won't be included in the output.

        Returns
        -------
        join_list : list
            A list of all the model classes or tablenames (depending on the
            value of ``format_out``) needed to connect all the elements in
            ``models``. The original elements in `models` are also included.

        """

        models = np.atleast_1d(models)
        format_out = format_out.lower()

        assert format_out in ['models', 'tables'], \
            'format_out must be either \'models\' or \'tables\'.'

        if len(models) == 0:
            raise ValueError('input list of models/tables to join is empty.')

        # Removes duplicates in input list.
        seen = set()
        seen_add = seen.add
        models = [xx for xx in models if not (xx in seen or seen_add(xx))]

        # Determines the type of input
        if isModel(models[0]):
            format_in = 'models'
        elif isinstance(models[0], (str, np.unicode_)):
            format_in = 'tables'
        else:
            raise ValueError('the format of the input list '
                             'cannot be understood.')

        # We create a list of the table paths for each input. We'll use
        # this list internally, and later format the output depending on
        # format_out.
        if format_in == 'models':
            tables = [self.getTablePath(model) for model in models]
        else:
            tables = models

        if nexus and isModel(nexus):
            nexus = self.getTablePath(nexus)
            assert nexus in self.graph.nodes(), \
                'nexus {0} is not a node in the model graph.'.format(nexus)

        for table in tables:
            assert table in self.graph.nodes(), \
                'table {0} is not a node in the model graph.'.format(table)

        if len(models) == 1:
            # Simple case in which we only have one table to join. We just
            # return the same table / model, depending on format_out
            # If nexus is defined, we recursively call getJoins().

            if not nexus:
                if format_out == 'tables':
                    return tables
                else:
                    return [self.graph.node[tables[0]]['model']]
            else:
                path = self.getJoins(models=[nexus, tables[0]], nexus=None,
                                     format_out=format_out)
                # Removes the nexus
                return path[1:]

        else:
            # We get all possible combinations of two elements in the input
            # list of models, and find the shortest path between them.
            # If nexus is defined, we get the joins between nexus and each item
            # in tables.

            joins = []

            if not nexus:
                for tableA, tableB in itertools.combinations(tables, r=2):
                    newJoins = self._getShortestPath(tableA, tableB,
                                                     format_out=format_out)
                    self._joinList(joins, newJoins)

            else:
                for tableB in tables:
                    newJoins = self._getShortestPath(nexus, tableB,
                                                     format_out=format_out,
                                                     removeA=True)
                    self._joinList(joins, newJoins)

            return joins

    def _joinList(self, listA, listB):
        """Joins two lists excluding duplicates. Join is made in place."""

        for table in listB:
            if table not in listA:
                listA.append(table)

        return listA

    def _getShortestPath(self, tableA, tableB, format_out='tables',
                         removeA=False):
        """Gets the shortest path between two nodes."""

        try:
            path = nx.shortest_path(self.graph, tableA, tableB)
        except nx.NetworkXNoPath:
            raise nx.NetworkXNoPath(
                'it is not possible to join tables {0} and {1}. '
                'Please, review your query.'.format(tableA, tableB))

        if removeA:
            path.remove(tableA)

        if format_out == 'tables':
            pathSet = path
        else:
            pathSet = [self.graph.node[table]['model'] for table in path]

        return pathSet
